import os
import json
# ------ Interaction Protocol -----------------------------------
from elsciRL.interaction_loops.standard_gym import GymInteractionLoop
# ------ Experiment Import --------------------------------------
from elsciRL.evaluation.standard_report import Evaluation
# ------ Agent Imports -----------------------------------------
# Universal Agents
from elsciRL.agents.agent_abstract import Agent, QLearningAgent
from elsciRL.agents.table_q_agent import TableQLearningAgent
from elsciRL.agents.DQN import DQNAgent
from elsciRL.agents.agent_abstract import Agent
# Gym based agents
from elsciRL.environment_setup.gym_translator import GymRegistration
from elsciRL.agents.stable_baselines.SB3_DQN import SB_DQN
from elsciRL.agents.stable_baselines.SB3_PPO import SB_PPO
from elsciRL.agents.stable_baselines.SB3_A2C import SB_A2C
from elsciRL.experiments.experiment_utils.env_manager import EnvManager
from elsciRL.experiments.experiment_utils.result_manager import ResultManager
from elsciRL.experiments.experiment_utils.agent_factory import AgentFactory
from elsciRL.experiments.training_procedures.default_exp_training import run_training_loop

# This is the main run functions for elsciRL to be imported
# Defines the train/test operators and imports all the required agents and experiment functions ready to be used
# The local main.py file defines the [adapters, configs, environment] to be input

# This should be where the environment is initialized and then episode_loop (or train/test) is run
# -> results then passed down to experiment to produce visual reporting (staticmethod)
# -> instruction following approach then becomes alternative form of this file to be called instead
# -> DONE: This means we have multiple main.py types (e.g. with/without convergence measure) so should create a directory and finalize naming for this

class GymExperiment:
    """This is the GYM Reinforcement Learning experiment setup for a flat of hierarchy agent. 
    - ONLY GYM AGENTS ARE SUPPORTED - To uses non-Gym agents use the StandardExperiment class
    - The agent is trained for a fixed number of episodes
    - Then learning is fixed to be applied during testing phase
    - Repeats (or seeds if environment start position changes) are used for statistical significant testing
    """
    def __init__(self, Config:dict, ProblemConfig:dict, Engine, Adapters:dict, save_dir:str, show_figures:str, window_size:float): 
        # Environment setup
        if isinstance(Engine, dict):
            print("\n Multiple Engines detected, will compare results across engines...")
            self.engine_comparison = True
            self.engine_list = Engine
        else:
            self.engine_comparison = False
            self.engine_list = {'DefaultEng':Engine}
        self.adapters = Adapters
        self.env_manager = EnvManager(GymInteractionLoop, Adapters)
        # ---
        # Configuration setup
        self.ExperimentConfig = Config
        self.LocalConfig = ProblemConfig
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)
        self.save_dir = save_dir
        self.show_figures = show_figures
        if (self.show_figures.lower() != 'y') and (self.show_figures.lower() != 'yes'):
            print("Figures will not be shown and only saved.")
        try:
            self.setup_info = self.ExperimentConfig['data'] | self.LocalConfig['data'] 
        except Exception:
            self.setup_info = self.ExperimentConfig | self.LocalConfig 
        self.training_setups: dict = {}
        self.trained_agents: dict = {}
        self.num_training_seeds = self.setup_info['number_training_seeds']
        self.test_agent_type = self.setup_info['test_agent_type']
        self.analysis = Evaluation(window_size=window_size)
        self.result_manager = ResultManager(self.analysis)
        self.agent_factory = AgentFactory(self.adapters, self.setup_info)
        self.reward_signal = None

    def add_agent(self, agent_name:str, agent):
        self.agent_factory.register_agent(agent_name, agent)
        print("\n Agent added to experiment, all available agents: ", self.agent_factory.agent_types)

    def train(self):
        if not os.path.exists(self.save_dir):
            os.mkdir(self.save_dir)
        for engine_name, engine in self.engine_list.items():
            for agent_type in self.setup_info['agent_select']:
                for adapter in self.setup_info['adapter_input_dict'][agent_type]:
                    train_setup_info = self.setup_info.copy()
                    agent_parameters = train_setup_info['agent_parameters'][agent_type]
                    train_setup_info['agent_type'] = agent_type
                    train_setup_info['agent_name'] = f"{engine_name}{agent_type}_{adapter}_{agent_parameters}"
                    train_setup_info['adapter_select'] = adapter
                    train_setup_info['train'] = True
                    number_training_repeats = self.ExperimentConfig["number_training_repeats"]
                    # Induce reward signal into engine
                    engine.reward_signal = self.reward_signal
                    # Use run_training_loop utility
                    trained_agents, seed_results_connection, temp_agent_store, training_results_stored, observed_states_stored = run_training_loop(
                        self.env_manager,
                        self.agent_factory,
                        self.result_manager,
                        training_render=False,
                        training_render_save_dir=None,
                        save_dir=self.save_dir,
                        engine_name=engine_name,
                        engine=engine,
                        agent_type=agent_type,
                        adapter=adapter,
                        all_adapters=self.adapters,
                        train_setup_info=train_setup_info,
                        trained_agents=self.trained_agents,
                        num_training_seeds=self.num_training_seeds,
                        test_agent_type=self.test_agent_type,
                        show_figures=self.show_figures,
                        number_training_repeats=number_training_repeats,
                        gym_env=True  # Use gym_env flag if available
                    )
                    self.trained_agents.update(trained_agents)
                    self.training_setups[f'Training_Setup_{engine_name}_{agent_type}_{adapter}'] = train_setup_info
        if (self.ExperimentConfig["number_training_repeats"] > 1) or (self.num_training_seeds > 1):
            self.result_manager.training_variance_report(self.save_dir, self.show_figures)
        return self.training_setups

    def test(self, training_setups:str=None):
        if training_setups is None:
            training_setups = self.training_setups
        else:
            training_setups = json.load(training_setups)
        for training_key in list(training_setups.keys()):    
            test_setup_info = training_setups[training_key]
            test_setup_info['train'] = False # Testing Phase
            test_setup_info['training_results'] = False
            test_setup_info['observed_states'] = False
            print("----------")
            print("Testing results for trained agents in saved setup configuration:")
            print(test_setup_info['train_save_dir'])
            number_training_repeats = test_setup_info['number_test_repeats']
            agent_adapter = test_setup_info['agent_type'] + "_" + test_setup_info['adapter_select']
            for engine_name, engine in self.engine_list.items():
                for testing_repeat in range(0, test_setup_info['number_test_repeats']):  
                    engine.reward_signal = None # clear instr reward signal
                    env = self.env_manager.create_gym_env(Engine=engine, Adapters=self.adapters, local_setup_info=test_setup_info)
                    start_obs = env.start_obs
                    goal = str(start_obs).split(".")[0] + "---GOAL"
                    print("Flat agent Goal: ", goal)
                    if goal in self.trained_agents.get(f"{engine_name}_{test_setup_info['agent_type']}_{test_setup_info['adapter_select']}", {}):
                        print("Trained agent available for testing.")
                        all_agents = self.trained_agents[str(engine_name) + '_' + test_setup_info['agent_type']+'_'+test_setup_info['adapter_select']][goal]
                        if not isinstance(all_agents, list):
                            all_agents = [all_agents]
                    else:
                        print("NO agent available for testing position.")
                    env.agent.epsilon = 0 # Remove random actions
                    for ag,agent in enumerate(all_agents):
                        env.results.reset() # Reset results table for each agent
                        env.start_obs = start_obs
                        env.agent = agent
                        env.agent.epsilon = 0 # Remove random actions
                        agent_adapter = test_setup_info['agent_type'] + "_" + test_setup_info['adapter_select']
                        # ---
                        # Testing generally is the agents replaying on the testing ENV
                        testing_results = env.episode_loop() 
                        test_save_dir = (self.save_dir+'/'+ str(engine_name) + '_' + agent_adapter + '__testing_results_' + str(goal).split("/")[0]+"_"+"agent"+str(ag)+"-repeat"+str(testing_repeat))
                        if not os.path.exists(test_save_dir):
                            os.mkdir(test_save_dir)
                        # Produce training report with Analysis.py
                        Return = self.analysis.test_report(testing_results, test_save_dir, self.show_figures)
        if (self.ExperimentConfig["number_training_repeats"] > 1) or (self.test_agent_type.lower() == 'all'):
            self.result_manager.testing_variance_report(self.save_dir, self.show_figures)


import os
import torch
from tqdm import tqdm
from torch import Tensor
import matplotlib.pyplot as plt
import string
# ------ Interaction Protocol -----------------------------------
from elsciRL.interaction_loops.standard import StandardInteractionLoop
from elsciRL.interaction_loops.state_search import episode_loop
from joblib import Parallel, delayed, cpu_count
# ------ Agent Imports -----------------------------------------
from elsciRL.agents.table_q_agent import TableQLearningAgent
from elsciRL.agents.DQN import DQNAgent
from elsciRL.agents.random_agent import RandomAgent
# --------------------------------------------------------------
# TODO use encoder defined by config not manual import
from elsciRL.encoders.language_transformers.MiniLM_L6v2 import LanguageEncoder

# Only use Qlearntab for for search agent for speeeeeed and randomn exploration
AGENT_TYPES = {
    "Qlearntab": TableQLearningAgent
}



class elsciRLSearch:
    def __init__(self, Config:dict, LocalConfig:dict, 
                 Engine, Adapters:dict,
                 save_dir:str, 
                 number_exploration_episodes:int = 10000,
                 match_sim_threshold:float=0.9,
                 observed_states:dict = None, observed_states_encoded:Tensor = None,
                 context_length:int=1000):
        
        # ----- Configs
        # Meta parameters
        self.ExperimentConfig = Config
        # Local Parameters
        self.ProblemConfig = LocalConfig
        
        self.engine = Engine
        self.adapters = Adapters
        self.env = StandardInteractionLoop
        self.setup_info:dict = self.ExperimentConfig | self.ProblemConfig  
        self.training_setups: dict = {}
        self.instruction_results:dict = {}
        
        self.save_dir = save_dir
        # Create save directory if it does not exist
        if not os.path.exists(self.save_dir):
            os.makedirs(self.save_dir)

        # Unsupervised search parameters
        self.agent_type: str = "Qlearntab" # fix to one type for fast search
        self.epsilon: float = 1 # fully random search agent
        self.number_exploration_episodes: int = number_exploration_episodes

        # Unsupervised search parameters
        self.context_length = context_length # Limit length of observed states
        self.enc = LanguageEncoder()
        self.sim_threshold: float = match_sim_threshold
        self.cos = torch.nn.CosineSimilarity(dim=1)  

        # Initialize observed states
        if (observed_states is not None) and (len(observed_states) > 0):
            self.observed_states = observed_states
        else:
            self.observed_states = {}
        # Encode all observed states 
        if (observed_states_encoded is not None) and (len(observed_states_encoded) > 0):
            self.str_states_encoded = observed_states_encoded
        else:
            if self.observed_states is None:
                self.str_states_encoded = None
            else:
                if len(self.observed_states) > 0:
                    print("\n -------------------------------\n")
                    str_states = [str_state[:self.context_length] for str_state in self.observed_states.values()]
                    print(f"Encoding {len(str_states)} observed states in {int(len(str_states)/256)} batches of size {256}.")
                    self.str_states_encoded = self.enc.encode(state=str_states, progress_bar=True)
        

    def search(self, action_cap:int=100):
        # Fixed to Tabular Q learning agent for now
        agent_type = self.agent_type
        # We are adding then overriding some inputs for exploration setups
        train_setup_info = self.setup_info.copy()
        # Override action cap for shorter term sub-goals for faster learning
        train_setup_info['training_action_cap'] = action_cap 
        # ----- State Adapter Choice
        for adapter in train_setup_info["adapter_select"]:
            if (("language" in adapter.lower()) |
                ("lang" in adapter.lower()) |
                ("_l" in adapter.lower())|
                ("l_" in adapter.lower())):
                train_setup_info["adapter_select"] = adapter
                break

        #adapter = train_setup_info["adapter_select"][0]
        # ----- Agent parameters
        agent_parameters = train_setup_info["agent_parameters"][agent_type]
        train_setup_info['agent_type'] = agent_type
        train_setup_info['agent_name'] = (str(agent_type) + '_' +
                                        str(adapter) + '_' +
                                        str(agent_parameters))
        train_setup_info['adapter_select'] = adapter
        # ----- init agent
        player = AGENT_TYPES[agent_type](**agent_parameters)
        train_setup_info['agent'] = player
        # -----
        # Set env function to training
        train_setup_info['train'] = True
        # --- 
        # Set exploration parameters
        train_setup_info['number_training_episodes'] = self.number_exploration_episodes 
        train_setup_info['epsilon'] = self.epsilon 
        # ---------------------------------elsciRL-----------------------------------------
        # Train on Live system for limited number of total episodes
        train_setup_info['training_results'] = False
        train_setup_info['observed_states'] = False
        train_setup_info['experience_sampling'] = False
        train_setup_info['live_env'] = True 
        # ---------------------------
        # Init environment to define current position
        train_setup_info['sub_goal'] = None 
        sample_env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=train_setup_info)
        print("Environment Init Position: ", sample_env.start_obs)
        # ---
        # Explore env with limited episodes
        # Environment now init here and called directly in experimental setup loop
        # - setup elsciRL info
        # Train on Live system for limited number of total episodes
        # Parallel processing for faster episode runs
        parallel = Parallel(n_jobs=cpu_count(True), prefer="processes", verbose=0)
        number_parallel_batches = 10
        if self.number_exploration_episodes > number_parallel_batches:
            number_episodes_per_parallel = int(self.number_exploration_episodes / number_parallel_batches)
        else:
            number_episodes_per_parallel
        observed_state_output = parallel(delayed(episode_loop)(Engine=self.engine, Adapters=self.adapters, 
                                                                local_setup_info=train_setup_info, 
                                                                number_episodes=number_episodes_per_parallel,
                                                                batch_number=i) for i in tqdm(range(number_parallel_batches)))
        for batch in observed_state_output:
            self.observed_states.update(batch)

        train_setup_info['training_results'] = None
        train_setup_info['observed_states'] = self.observed_states
        train_setup_info['experience_sampling'] = None
        # Extract visited states from env
        # ---------------------------
                
        return self.observed_states
    
    def match(self, action_cap:int=5, instructions:list=[''], instr_descriptions:list=['']):
        device = "cuda" if torch.cuda.is_available() else "cpu" 
        # Fixed to Tabular Q learning agent for now
        agent_type = self.agent_type
        # We are adding then overriding some inputs for exploration setups
        train_setup_info = self.setup_info.copy()
        # Override action cap for shorter term sub-goals for faster learning
        train_setup_info['training_action_cap'] = action_cap 
        # ----- State Adapter Choice
        adapter = train_setup_info["adapter_select"][0]
        self.agent_adapter = agent_type+'_'+adapter
        # ----- Agent parameters
        agent_parameters = train_setup_info["agent_parameters"][agent_type]
        train_setup_info['agent_type'] = agent_type
        train_setup_info['agent_name'] = (str(agent_type) + '_' +
                                        str(adapter) + '_' +
                                        str(agent_parameters))
        train_setup_info['adapter_select'] = adapter
        # ----- init agent
        player = AGENT_TYPES[agent_type](**agent_parameters)
        train_setup_info['agent'] = player
        # -----
        # Set env function to training
        train_setup_info['train'] = True
        # ---------------------------------elsciRL-----------------------------------------
        # Train on Live system for limited number of total episodes
        train_setup_info['training_results'] = False
        train_setup_info['observed_states'] = False
        train_setup_info['experience_sampling'] = False
        train_setup_info['live_env'] = True 
        # ---------------------------
        # Init environment to define current position
        train_setup_info['sub_goal'] = None 
        sample_env = self.env(Engine=self.engine, Adapters=self.adapters, local_setup_info=train_setup_info)
            
        if self.str_states_encoded is None:
            str_states = [str_state[:self.context_length] for str_state in self.observed_states.values()]
            self.str_states_encoded = self.enc.encode(str_states)
        # New: user input here
        #instructions, instr_descriptions = self.elsciRL_input.user_input()
        # DEMO SETS THIS AS FUNCTION INPUT
        # ---------------------------
        best_match_results = {}
        for i,instr in enumerate(instructions):
            if i == 0:  
                # env start is '0' so force rest to start from '1' instead otherwise 
                # breaks uniqueness requirement for instructions
                instruction = str(sample_env.start_obs).split(".")[0] + "---" + str(int(instr)+1)
            else:
                instruction = str(int(instructions[i-1])+1) + "---" + str(int(instr)+1)
            instr_description = instr_descriptions[i]
            print(f"\nFinding match for instruction: {instruction}")
            
            if (type(instr_description)) == type(''):
                if (len(instr_description.split('.')) > 1):
                    instr_description = instr_description.split('.')
                    # Remove punctuation from description
                    instr_description = [s.translate(str.maketrans('', '', string.punctuation)) for s in instr_description]
                    # Remove empty strings from description
                    instr_description = [s.strip() for s in instr_description if s.strip()]
                    instr_description = ' '.join(instr_description)
                else:
                    instr_description = [instr_description.translate(str.maketrans('', '', string.punctuation)).strip()]
            # Create tensor vector of description
            instruction_vector = self.enc.encode(instr_description)
            if instruction_vector.size()[0] > 1:
                print(f"Instruction vector size: {instruction_vector.size()}")
                print(f"\n Instruction vector: {instr_description} \n")
            # Default fedeback layer - DEMO wont currently updated this 
            feedback_layer = torch.zeros(instruction_vector.size()).to(device)
            # EXPLORE TO FIND LOCATION OF SUB-GOAL
            sub_goal = None
            # ---------------------------
            if (instruction in self.instruction_results):
                if (self.agent_adapter) in self.instruction_results[instruction]:
                    # We use feedback layer even if sub_goal not a good match
                    feedback_layer = self.instruction_results[instruction][self.agent_adapter]['feedback_layer']
                    if (self.instruction_results[instruction][self.agent_adapter]['sim_score']>=self.sim_threshold):
                        sub_goal = self.instruction_results[instruction][self.agent_adapter]['sub_goal'][0]
                        sub_goal_list = self.instruction_results[instruction][self.agent_adapter]['sub_goal']
                        sim = self.instruction_results[instruction][self.agent_adapter]['sim_score']
                else:
                    self.instruction_results[instruction][self.agent_adapter] = {}
                self.instruction_results[instruction][self.agent_adapter]['count'] = self.instruction_results[instruction][self.agent_adapter]['count']+1
            else:
                self.instruction_results[instruction] = {}    
                self.instruction_results[instruction]['instr_description'] = instr_description
                self.instruction_results[instruction][self.agent_adapter] = {} 
                self.instruction_results[instruction][self.agent_adapter]['count'] = 1
                self.instruction_results[instruction][self.agent_adapter]['action_cap'] = action_cap
                self.instruction_results[instruction][self.agent_adapter]['feedback_layer'] = feedback_layer
            # ---------------------------
            while not sub_goal:
                # If no description -> no sub-goal (i.e. envs terminal goal position)
                if not instr_description: 
                    sub_goal = None
                    # If no sub-goal -> find best match of description from env 
                else:
                    # Compare to instruction vector                            
                    max_sim = -1
                    # all states that are above threshold 
                    sub_goal_list = []
                    for obs_state_idx,obs_state in tqdm(enumerate(self.observed_states), total=len(self.observed_states)):
                        #str_state = self.observed_states[obs_state][:self.context_length]
                        t_state = self.str_states_encoded[obs_state_idx] #self.enc.encode(str_state)
                        # ---
                        sim = self.cos(torch.add(t_state, feedback_layer), instruction_vector).mean()
                       
                        if sim > max_sim:
                            max_sim  = sim
                            sub_goal_max = obs_state
                            #sub_goal_max_t = t_state
                        if sim >= self.sim_threshold:
                            sub_goal = obs_state # Sub-Goal code
                            sub_goal_max = obs_state
                            #sub_goal_t = t_state
                            sub_goal_list.append(sub_goal)
    
                    # OR if none above threshold matching max sim
                    # TODO: IMPROVE RUNTIME HERE BY NOT RUNNING SEARCH TWICE
                    if max_sim < self.sim_threshold:
                        print(" --- Best match not above threshold, finding highest sim instead...")
                        sub_goal = sub_goal_max
                        #sub_goal_t = sub_goal_max_t                                    
                        # Find all states that have same sim as max
                        for obs_state_idx,obs_state in tqdm(enumerate(self.observed_states), total=len(self.observed_states)):
                            #str_state = self.observed_states[obs_state][:self.context_length]
                            t_state = self.str_states_encoded[obs_state_idx] #self.enc.encode(str_state)
                            # ---
                            sim = self.cos(torch.add(t_state, feedback_layer), instruction_vector).mean()
                            if sim >= (max_sim):
                                sub_goal_list.append(obs_state)

                    if max_sim < self.sim_threshold:
                        print("Minimum sim for observed states to match instruction not found, using best match instead. Best match sim value = ", max_sim )
                # If adapter is poor to match to instruction vector none of them observed states match
                if (max_sim<-1)|(sub_goal is None):#|(max_sim>1):
                    print("All observed states result in similarity outside bounds (i.e. not found).")
                    print("Max sim found = ", max_sim)
                    sub_goal_list = ['']
                    break

            if (self.agent_adapter) in self.instruction_results[instruction]:
                self.instruction_results[instruction][self.agent_adapter]['sub_goal'] = sub_goal_list
                self.instruction_results[instruction][self.agent_adapter]['sim_score'] = sim
                self.instruction_results[instruction][self.agent_adapter]['feedback_layer'] = feedback_layer
            else:
                self.instruction_results[instruction][self.agent_adapter] = {}
                self.instruction_results[instruction][self.agent_adapter]['count'] = 1
                self.instruction_results[instruction][self.agent_adapter]['action_cap'] = action_cap
                self.instruction_results[instruction][self.agent_adapter]['sub_goal'] = sub_goal_list
                self.instruction_results[instruction][self.agent_adapter]['sim_score'] = sim
                self.instruction_results[instruction][self.agent_adapter]['feedback_layer'] = feedback_layer

            best_match_results[instruction] = {}
            best_match_results[instruction]['sub_goal'] = self.observed_states[sub_goal]
            best_match_results[instruction]['best_match'] = sub_goal_max
        
        # Store data for feedback and plot
        self.instructions = instructions
        self.instr_descriptions = instr_descriptions
                       
        return best_match_results, self.instruction_results
    
    def feedback(self, feedback_type:str='positive', feedback_increment:float=0.5,
                 plot:bool=False, plot_save_dir:str=None):
        # Update feedback layer to based on validation
        # Get instruction vector
        instructions = self.instructions
        instr_descriptions = self.instr_descriptions

        sim_total_delta = 0
        for i,instr_description in enumerate(instr_descriptions):
            instruction = list(self.instruction_results.keys())[i]
            print(f"\nFinding match for instruction: {instruction}")
            
            if (type(instr_description)) == type(''):
                if (len(instr_description.split('.')) > 1):
                    instr_description = instr_description.split('.')
                    # Remove punctuation from description
                    instr_description = [s.translate(str.maketrans('', '', string.punctuation)) for s in instr_description]
                    # Remove empty strings from description
                    instr_description = [s.strip() for s in instr_description if s.strip()]
                    instr_description = ' '.join(instr_description)
                else:
                    instr_description = [instr_description.translate(str.maketrans('', '', string.punctuation)).strip()]
            # Create tensor vector of description
            instruction_vector = self.enc.encode(instr_description)
            # Get sub-goal vector
            sub_goal = self.instruction_results[instruction][self.agent_adapter]['sub_goal'][0]
            sub_goal_t = self.enc.encode(sub_goal.replace(".", "")) # We cannot parse multiple sentences from sub-goal for feedback layer update
            # Get feedback layer vector
            feedback_layer = self.instruction_results[instruction][self.agent_adapter]['feedback_layer']
            try:
                if feedback_type == 'positive':
                    # Positive feedback - add to feedback layer
                    feedback_layer = torch.add(feedback_layer, feedback_increment*(torch.sub(instruction_vector, sub_goal_t))) 
                elif feedback_type == 'negative':
                    feedback_layer = torch.sub(feedback_layer, feedback_increment*(torch.sub(instruction_vector, sub_goal_t)))
                else:
                    raise ValueError(f"Invalid feedback type: {feedback_type}")
            except:
                print(f"Error updating feedback layer for instruction: {instruction}")
                print(f"Feedback type: {feedback_type}")
                print(f"Feedback increment: {feedback_increment}")
                print(f"Instruction vector: {instruction_vector.shape} - {instruction_vector.size()}")
                print(f"Sub-goal vector: {sub_goal_t.shape} - {sub_goal_t.size()}")
                print(f"Feedback layer: {feedback_layer.shape} - {feedback_layer.size()}")
                
            self.instruction_results[instruction][self.agent_adapter]['feedback_layer'] = feedback_layer
            # Average sim across each sentence in instruction vs state
            sim = self.cos(torch.add(sub_goal_t, feedback_layer), instruction_vector).mean() # bounds: [0,1]       
            sim_delta = sim-self.instruction_results[instruction][self.agent_adapter]['sim_score']
            print(f"--- Change in sim results with {feedback_type} reinforcement of correct state match = {sim_delta:.2f}")
            print(f"--- New Sim Value = {sim:.2f}") 

            if plot:
                if plot_save_dir is not None:
                    self.file_name = self.save_dir + '/' + plot_save_dir + '/feedbackplot__' + instruction[:4] + '.png'
                else:
                    self.file_name = self.save_dir + '/feedbackplot__'  + instruction[:4] + '.png'

                print(f"--- Plotting feedback for instruction: {instruction} to {self.file_name}")
                return self.sim_plot(sim_delta, self.instruction_results[instruction][self.agent_adapter]['sim_score'], feedback_type)

            self.instruction_results[instruction][self.agent_adapter]['sim_score'] = sim
            sim_total_delta += sim_delta

        sim_delta_avg = sim_total_delta / len(instructions)
        return sim_delta_avg
    
    def sim_plot(self, sim_delta:float, sim:float, feedback_type:str='positive'):
        # Plot sim delta and sim value
        if type(sim) == torch.Tensor:
            sim = sim.item()

        # Create arbitrary 2D vectors for visualization
        instruction_vec = [1, 1]  # Unit vector along x-axis
        
        # Calculate predicted vector based on similarity (sim)
        angle = torch.arccos(torch.tensor(sim)) # Convert sim to angle
        pred_x = torch.cos(angle)
        pred_y = torch.sin(angle)
        pred_vec = [pred_x.item(), pred_y.item()]
        
        # Calculate updated vector after feedback
        if pred_x > instruction_vec[0]:
            new_angle = torch.arccos(torch.tensor(sim - sim_delta))
        else:
            new_angle = torch.arccos(torch.tensor(sim + sim_delta))
      
        new_x = torch.cos(new_angle)
        new_y = torch.sin(new_angle)
        new_vec = [new_x.item(), new_y.item()]

        # Plot vectors
        fig = plt.figure(figsize=(8,8))
        plt.quiver(0, 0, instruction_vec[0], instruction_vec[1], angles='xy', scale_units='xy', scale=1, color='black', label='Instruction')
        plt.quiver(0, 0, pred_vec[0], pred_vec[1], angles='xy', scale_units='xy', scale=1, color='grey', linestyle='dashed', facecolor='none', linewidth=2,
          width=0.0001, headwidth=300, headlength=500, label='Initial Prediction')
        if sim_delta >= 0:
            plt.quiver(0, 0, new_vec[0], new_vec[1], angles='xy', scale_units='xy', scale=1, color='blue', label='After Feedback')
        else:
            # For negative feedback, use a different color
            plt.quiver(0, 0, new_vec[0], new_vec[1], angles='xy', scale_units='xy', scale=1, color='red', label='After Feedback')
        
        plt.xlim(-1.5, 1.5)
        plt.ylim(-1.5, 1.5)
        plt.grid(True)
        plt.axhline(y=0, color='k', linestyle='-', alpha=0.3)
        plt.axvline(x=0, color='k', linestyle='-', alpha=0.3)
        plt.title(f'Vector Similarity After {feedback_type.title()} Feedback\nUpdated Sim: {abs(sim+sim_delta):.2f}, Delta: {sim_delta:.2f}')
        plt.legend()
        plt.tight_layout()
        plt.savefig(self.file_name, dpi=100)
        plt.close()

        return self.file_name


import numpy as np
import pandas as pd
import os 
import json

def combined_tabular_analysis_results(results_dir:str='', analysis_type='training'):
    if results_dir == '':
        raise ValueError("Save directory not specified.")
    analysis_type = analysis_type.lower() # lowercase analysis type input
    # Get sub-dir for each problem-experiment type
    instruction_folders = [os.path.join(results_dir, instr) for instr in os.listdir(results_dir) if os.path.isdir(os.path.join(results_dir, instr))]
    variance_results = {}
    for instr_folder_dir in instruction_folders:
        instr_id = instr_folder_dir.split('/')[-1].split('//')[-1].split('\\')[-1].split('\\\\')[-1]
        if instr_id not in variance_results.keys():
            variance_results[instr_id] = {}
        print(f"Processing {instr_id} for {analysis_type} analysis.")
        problem_folders = [name for name in os.listdir(instr_folder_dir) if os.path.isdir(os.path.join(instr_folder_dir, name))]
        # Find experiment folders
        # - Capture case where there is only one experiment type
        # and so wont have sub-directory for experiments to search
        for experiment_dir in problem_folders:
            if analysis_type == 'training':
                experiment_name = experiment_dir+'_training'
                file_names = [name for name in os.listdir(instr_folder_dir+'/'+experiment_dir) if name[0:25] == 'training_variance_results']
            elif analysis_type == 'testing':
                experiment_name = experiment_dir+'_testing'
                file_names = [name for name in os.listdir(instr_folder_dir+'/'+experiment_dir) if name[0:24] == 'testing_variance_results']
            else:
                raise ValueError("Analysis type must be either 'training' or 'testing'.")

            if experiment_name not in variance_results[instr_id].keys():
                variance_results[instr_id][experiment_name] = {}

            for file in file_names:
                results = pd.read_csv(instr_folder_dir+'/'+experiment_dir+'/'+file)
                agent = results['agent'].iloc[0].split('__')[0]
                if agent not in variance_results[instr_id][experiment_name].keys():
                    variance_results[instr_id][experiment_name][agent] = {}
                
                # Calculate Mean and Std Dev
                variance_results[instr_id][experiment_name][agent]['num_repeats'] = results['num_repeats'].iloc[0]
                variance_results[instr_id][experiment_name][agent]['number_episodes'] = results.index.max() + 1
                # - rolling avg R per episode
                variance_results[instr_id][experiment_name][agent]['mean'] = results['avg_R_mean'].mean()
                variance_results[instr_id][experiment_name][agent]['median'] = results['avg_R_mean'].median()
                variance_results[instr_id][experiment_name][agent]['std_error'] = results['avg_R_mean'].sem()
                variance_results[instr_id][experiment_name][agent]['std_dev'] = results['avg_R_mean'].std()
                variance_results[instr_id][experiment_name][agent]['variance'] = results['avg_R_mean'].var()
                # - cumulative R per episode
                variance_results[instr_id][experiment_name][agent]['cum_R_mean'] = results['cum_R_mean'].mean()
                variance_results[instr_id][experiment_name][agent]['cum_R_median'] = results['cum_R_mean'].median()
                variance_results[instr_id][experiment_name][agent]['cum_R_std_error'] = results['cum_R_mean'].sem()
                variance_results[instr_id][experiment_name][agent]['cum_R_std_dev'] = results['cum_R_mean'].std()
                variance_results[instr_id][experiment_name][agent]['cum_R_variance'] = results['cum_R_mean'].var()
                # - time avg per episode
                variance_results[instr_id][experiment_name][agent]['time_avg'] = results['time_mean'].mean()

    variance_results_df = pd.DataFrame.from_dict(
        {f"{instr}/{experiment}/{agent}": data for instr, experiments in variance_results.items() 
         for experiment, agents in experiments.items() 
         for agent, data in agents.items()},
        orient='index'
    ).reset_index()
    variance_results_df.columns = ['Instruction/Experiment/Agent', 'Num Repeats', 'Number Episodes', 
                                    'Avg R Mean', 'Avg R Median', 'Avg R Std Error', 'Avg R Std Dev', 'Avg R Variance',
                                    'Cumulative R Mean', 'Cumulative R Median', 'Cumulative R Std Error',
                                    'Cumulative R Std Dev', 'Cumulative R Variance', 'Time Avg']
    # Save the combined results to a CSV file
    combined_results_filename = f"{analysis_type}_combined_results.csv"
    combined_results_path = os.path.join(results_dir, combined_results_filename)
    variance_results_df.to_csv(combined_results_path, index=False)
    


import os

def render_current_result(training_setup, current_environment, current_agent, local_save_dir):
    """Apply fixed policy to render current decision making for limited number of episodes."""
    # Override input training setups with previously saved 
    
    test_setup_info = training_setup.copy()

    test_setup_info['train'] = False # Testing Phase
    test_setup_info['training_results'] = False
    test_setup_info['observed_states'] = False
    test_setup_info['experience_sampling'] = False
    print("----------")
    print("Rendering trained agent's policy:")

    env = current_environment
    # ---
    env.number_episodes = 1 # Only render 1 episode
    env.agent = current_agent
    env.agent.epsilon = 0 # Remove random actions
    # ---
    # Render results
    if not os.path.exists(local_save_dir):
        os.mkdir(local_save_dir)
    env.episode_loop(render=True, render_save_dir=local_save_dir) 


import ollama
import json
import re
from typing import List, Dict, Optional, Union
import random
import logging

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class OllamaTaskBreakdown:
    """
    Ollama LLM chat class for breaking down complex tasks into manageable sub-goals.
    """
    
    def __init__(self, model_name: str = "llama3.2", context_length: int = 4000, host: str = "localhost:11434",
                 input_prompt: str = None, observed_states: dict = None):
        """
        Initialize the Ollama chat client.
        
        Args:
            model_name: Name of the Ollama model to use
            context_length: Maximum number of messages to keep in context
            host: Ollama server host and port
        """
        self.model_name = model_name.strip().lower()
        self.host = host
        self.client = ollama.Client(host=host)
        self.conversation_history = []
        self.context_length = context_length
        
        # Verify model availability
        try:
            self._verify_model()
        except Exception as e:
            logger.error(f"Failed to initialize Ollama client: {e}")
            raise

        # Set input prompt if provided
        if input_prompt is not None:
            logger.info(f"Using input prompt: {input_prompt[:self.context_length]}...")  # Log first 100 chars
            self.input_prompt = input_prompt

        # Get randomly sampled observed states if provided to use for prompt output structure
        if observed_states is not None:
            print("Using observed states for prompt output structure.")
            keys = random.sample(list(observed_states), min(10, len(observed_states)))
            self.observed_states_prompt = str([observed_states[k] for k in keys])
    
    def _verify_model(self):
        """Verify that the specified model is available."""
        try:
            models = self.client.list()
            available_models = [model['model'] for model in models['models']]
            available_models_short = [model['model'].split(':')[0] for model in models['models']]
            if (self.model_name.strip not in available_models) and (self.model_name not in available_models_short):
                logger.warning(f"Model {self.model_name} not found. Available models: {available_models}")
            else:
                logger.info(f"Model {self.model_name} is available.")
        except Exception as e:
            logger.error(f"Error verifying model: {e}")
            raise
    
    def chat(self, prompt: str, system_prompt: str = None) -> str:
        """
        Send a chat message to the Ollama model.
        
        Args:
            prompt: User prompt/message
            system_prompt: Optional system prompt to set context
            
        Returns:
            Model response as string
        """
        try:
            messages = []
            
            # Add system prompt if provided
            if system_prompt:
                messages.append({
                    'role': 'system',
                    'content': system_prompt
                })
            
            # Add conversation history
            messages.extend(self.conversation_history)
            
            # Add current user message
            messages.append({
                'role': 'user',
                'content': prompt
            })
            
            response = self.client.chat(
                model=self.model_name,
                messages=messages[:self.context_length]
            )
            
            assistant_response = response['message']['content']
            
            # Update conversation history
            self.conversation_history.append({
                'role': 'user',
                'content': prompt
            })
            self.conversation_history.append({
                'role': 'assistant',
                'content': assistant_response
            })
            
            return assistant_response
            
        except Exception as e:
            logger.error(f"Error in chat: {e}")
            raise
    
    def break_down_task(self, task_prompt: str, max_subgoals: int = 5) -> List[str]:
        """
        Break down a complex task into sub-goals using the LLM.
        
        Args:
            task_prompt: The main task to break down
            max_subgoals: Maximum number of sub-goals to generate
            
        Returns:
            List of sub-goal strings
        """
        system_prompt = f"""You are an expert task planner. Your job is to break down complex tasks into smaller, manageable sub-goals.
Your instructions will be used to help an agent that interacts with the environment.
You do not interact with the environment directly, but your output will be used to plan actions.
DO NOT return any steps that are not actionable, exclude any 'gather information' 'generate' 'think', 'assess', 'look around', 'determine', 'research', 'analyze', 'review', 'explore', 'investigate', 'identify', 'understand', 'examine', 'evaluate', 'plan', 'decide'.
You must generate actionable steps based on the task description provided that an agent can complete.
INSTRUCTIONS:
1. Analyze the given task and break it down into {max_subgoals} or fewer concrete sub-goals
2. Each sub-goal should be specific, actionable, and measurable
3. List each sub-goal on a separate line, numbered (1., 2., 3., etc.)
4. Focus on creating logical, sequential sub-goals that when completed will achieve the main objective
5. Do not include any additional commentary or explanations, just the sub-goals

Example format:
1. First sub-goal description
2. Second sub-goal description
3. Third sub-goal description"""
        
        # If input prompt is provided, include it in the system prompt
        if hasattr(self, 'input_prompt'):
            system_prompt += f"\n\nThe problem's data source are provided by the following code, use this as guidance on how to act: {self.input_prompt}"

        # If observed states are provided, include them in the system prompt
        if hasattr(self, 'observed_states_prompt'):
            system_prompt += f"\n\nMatch your output language form to the observed language from the environment, here are some examples of the language structure:\n{self.observed_states_prompt}"

        user_prompt = f"""Task to break down: {task_prompt}

Please analyze this task and break it down into manageable sub-goals. Consider:
- What are the key components needed?
- What is the logical sequence of steps?
- Sub-goals must be defined sequentially?

Provide your response as a numbered list of sub-goals."""

        try:
            response = self.chat(user_prompt, system_prompt)
            return self._parse_subgoals_response(response, max_subgoals)
            
        except Exception as e:
            logger.error(f"Error breaking down task: {e}")
            raise
    
    def _parse_subgoals_response(self, response: str, max_subgoal: int = None) -> List[str]:
        """
        Parse the LLM response and extract sub-goals.
        
        Args:
            response: Raw LLM response containing numbered list
            
        Returns:
            List of sub-goal strings
        """
        try:
            sub_goals = []
            lines = response.strip().split('\n')
            
            for line in lines:
                line = line.strip()
                # Look for numbered items (1., 2., etc.) or bullet points
                if re.match(r'^\d+\.', line) or line.startswith('- ') or line.startswith('• '):
                    # Remove the numbering/bullet and clean up
                    clean_goal = re.sub(r'^\d+\.\s*', '', line)
                    clean_goal = re.sub(r'^[-•]\s*', '', clean_goal)
                    if clean_goal:
                        sub_goals.append(clean_goal.strip())
            
            # If no numbered items found, try to split by sentences or lines
            if not sub_goals:
                potential_goals = [line.strip() for line in lines if line.strip()]
                sub_goals = [goal for goal in potential_goals if len(goal) > 10]  # Filter out very short lines

            # Filter out non-actionable sub-goals
            actionable_sub_goals = []
            for goal in sub_goals:
                # Lowercase for easier matching
                goal_lower = goal.lower()
                # Filter out goals that are not actionable
                if any(word in goal_lower for word in ['think', 'consider', 'assess', 'look around', 'determine', 'research', 'analyze', 'review', 'explore', 'investigate', 'identify', 'understand', 'examine', 'evaluate', 'plan', 'decide']):
                    continue
                actionable_sub_goals.append(goal)

            # If we have more or fewer than max_subgoal, use LLM to summarise into exactly max_subgoal actionable steps
            summarise_prompt = (
                f"Given the following actionable sub-goals:\n"
                + "\n".join([f"- {g}" for g in actionable_sub_goals])
                + f"\n\nSummarise and combine them into exactly {max_subgoal} clear, actionable, and non-overlapping steps. "
"The steps are to interact with the environment and the final output used to plan, not to think, refine or plan or as inputs for a model so do not include text like 'generate' or 'produce'."
"Do not include any steps about thinking, assessing, looking around, creating a plan, or planning. "
"Return only a numbered list of actionable steps."
"Do not include any additional commentary or explanations, just the sub-goals. "
"Do not include any text that is not a sub-goal, such as 'Here are the sub-goals:' or 'The sub-goals are:' or 'These are the steps to take:' or 'Write an instruction:'. "
            )
            try:
                summary_response = self.chat(summarise_prompt)
                actionable_sub_goals = []
                lines = summary_response.strip().split('\n')
                for line in lines:
                    line = line.strip()
                    if re.match(r'^\d+\.', line):
                        clean_goal = re.sub(r'^\d+\.\s*', '', line)
                        if clean_goal:
                            actionable_sub_goals.append(clean_goal.strip())
                # If still not exactly max_subgoal, trim or pad
                if len(actionable_sub_goals) > max_subgoal:
                    actionable_sub_goals = actionable_sub_goals[:max_subgoal]
                elif len(actionable_sub_goals) < max_subgoal:
                    # Pad with generic actionable step if needed
                    actionable_sub_goals += [f"Complete the task step {i+1}" for i in range(len(actionable_sub_goals), max_subgoal)]
            except Exception as e:
                logger.error(f"Error summarising sub-goals with LLM: {e}")
                # Fallback: trim or pad
                if len(actionable_sub_goals) > max_subgoal:
                    actionable_sub_goals = actionable_sub_goals[:max_subgoal]
                elif len(actionable_sub_goals) < max_subgoal:
                    actionable_sub_goals += [f"Complete the task step {i+1}" for i in range(len(actionable_sub_goals), max_subgoal)]

            sub_goals = actionable_sub_goals

            # NOTE: HARD LIMIT ON NUMBER OF SUB-GOALS
            if max_subgoal is None or max_subgoal > len(sub_goals):
                max_subgoal = len(sub_goals)
            sub_goals = sub_goals[:max_subgoal]
            
            return sub_goals if sub_goals else [f"Complete the task: {response[:self.context_length]}..."]
            
        except Exception as e:
            logger.error(f"Error parsing sub-goals response: {e}")
            logger.error(f"Raw response: {response}")
            # Return a fallback sub-goal
            return [f"Complete the task: {response[:self.context_length]}..."]
    
    
    def clear_conversation(self):
        """Clear the conversation history."""
        self.conversation_history = []


# Example usage and testing
if __name__ == "__main__":
    # Example usage
    try:
        # Initialize the task breakdown system
        task_breaker = OllamaTaskBreakdown(model_name="llama3.2")
        
        # Example task
        example_task = "Build a machine learning model to predict customer churn for an e-commerce website"
        
        print(f"Breaking down task: {example_task}")
        print("-" * 50)
        
        # Break down the task
        sub_goals = task_breaker.break_down_task(example_task, max_subgoals=5)
        
        # Display the sub-goals
        print("Sub-goals:")
        for i, goal in enumerate(sub_goals, 1):
            print(f"{i}. {goal}")
        
    except Exception as e:
        print(f"Error in example: {e}")
        print("Make sure Ollama is running and the specified model is available.")



from abc import ABC, abstractmethod
from typing import List, Any
import os

from openai import OpenAI

try:
    from torch import Tensor
    from elsciRL.encoders.language_transformers.MiniLM_L6v2 import LanguageEncoder
except ImportError:
    print("Warning: torch or LanguageEncoder not found. Please ensure elsciRL is properly installed.")
    Tensor = None
    LanguageEncoder = None

class LLMAdapter(ABC):
    """Convert a general prompt and raw text state into a description of the state."""
    def __init__(self, base_prompt: str):
        super().__init__()
        # Define the fields that describe the state features:
        self.base_prompt = base_prompt

    @abstractmethod
    def _read(raw_state) -> list:
        # Read the data.
        # fill in the feature fields
        raise NotImplementedError


class GPTAdapter(LLMAdapter):
    """Adapter for OpenAI GPT models."""
    
    def __init__(self, base_prompt: str, model_name: str = "gpt-4"):
        super().__init__(base_prompt)
        self.model_name = model_name
        
        # Initialize the language encoder for encoding functionality
        if LanguageEncoder is not None:
            self.encoder = LanguageEncoder()
        else:
            print("Warning: LanguageEncoder not available. Encoding will not work.")
            self.encoder = None
    
    def _read(self, raw_state) -> list:
        """Read the data and fill in the feature fields."""
        # This method should be implemented based on specific requirements
        # For now, returning the raw state as a list
        return [raw_state] if isinstance(raw_state, str) else raw_state
    
    def call_gpt_api(self, prompt: str):
        """Call the OpenAI GPT API with the given prompt."""
        try:
            api_key = os.environ.get("OPENAI_API_KEY")
            if not api_key:
                raise ValueError("OPENAI_API_KEY environment variable not set")
            
            client = OpenAI(api_key=api_key)
            response = client.chat.completions.create(
                model=self.model_name,
                messages=[
                    {"role": "system", "content": self.base_prompt},
                    {"role": "user", "content": prompt}
                ],
                max_tokens=5000
            )
            return response.to_dict() if hasattr(response, 'to_dict') else response
        except Exception as e:
            print(f"Error calling OpenAI API: {e}")
            return None

    def process_gpt_response(self, response):
        """Process the response from OpenAI API."""
        if response and 'choices' in response:
            return response['choices'][0]['message']['content']
        return None

    def adapter(self, state: any, legal_moves: list = None, episode_action_history: list = None, encode: bool = True, indexed: bool = False) -> Tensor:
        """Returns the adapted form, may require input flag for encoded or non-encoded output."""
        # Build the full context prompt including legal moves and action history
        context_parts = []
        
        # Add state information
        if state is not None:
            context_parts.append(f"Current state: {state}")
        
        # Add legal moves if provided
        if legal_moves is not None and len(legal_moves) > 0:
            context_parts.append(f"Legal moves: {legal_moves}")
        
        # Add action history if provided
        if episode_action_history is not None and len(episode_action_history) > 0:
            recent_actions = episode_action_history[-5:]  # Last 5 actions
            context_parts.append(f"Recent actions: {recent_actions}")
        
        # Combine all context into a single prompt
        full_prompt = " | ".join(context_parts)
        
        # Get GPT response
        adapted_state = self.call_gpt_api(full_prompt)
        processed_response = self.process_gpt_response(adapted_state)
        
        if processed_response is None:
            processed_response = str(state) if state is not None else "No state available"
        
        # Handle encoding
        if encode:
            if self.encoder is not None:
                # Use the LanguageEncoder to encode the response
                state_encoded = self.encoder.encode(
                    state=processed_response,
                    legal_actions=legal_moves,
                    episode_action_history=episode_action_history,
                    indexed=indexed
                )
                return state_encoded
            else:
                print("Warning: Encoder not available, returning processed response as string")
                return processed_response
        else:
            return processed_response
        
    def sample(self, state: any):
        """Returns a sample of an adapted state form (typically initial position of the environment)."""
        if not state:
            state = 'The current state is empty.'
        return self.adapter(state, encode=True)

        



elsciRL_base_prompt = """
You are a helpful assistant that needs to describe the current state of a reinforcement learning environment to help an agent understand the context of the problem and how to act optimally.

The state can be text but is typically a list of numbers, you will be provided with prior actions and their outcome states and should use this information to describe the current state.

If no actions are provided, you should still describe the current state as best as you can.

You will be provided with a list of legal actions that the agent can take in the current state, you should describe these actions in a way that is useful for the agent to understand what it can do.

You do not need to provide any details about what the agent should do,  just describe the current state and the legal actions available to the agent in a single paragraph with less than 200 words.


"""


class DefaultAgentConfig:
    def __init__(self):
        self.data ={   
            "name": "Default",
            "problem_type": "Default",

            "instruction_chain": True,
            "instruction_chain_how": "continuous",
                         
            "number_training_episodes": 1000,
            "number_training_repeats": 5,
            "number_training_seeds": 1,

            "test_agent_type":"best",
            "number_test_episodes": 200,
            "number_test_repeats": 10,

            "agent_select": ["Qlearntab"],
            "adapter_select": ["default"],
            "agent_parameters":{
                "Qlearntab":{
                    "alpha": 0.1,
                    "gamma": 0.95,
                    "epsilon": 1,
                    "epsilon_step":0
                    },
                "DQN":{
                    "learning_rate": 0.001,
                    "gamma": 0.99,
                    "epsilon": 1.0,
                    "epsilon_min": 0.01,
                    "epsilon_decay": 0.995,
                    "memory_size": 10000,
                    "batch_size": 64,
                    "target_update": 10,
                    "hidden_size": 128
                },
            }
        }



#  Define data through class function so it can be called within package
# Instead of using a .json file which is hard to load from local install
# NOTE: MAKE SURE TO TRUST REPOSITORIES BEFORE RUNNING CODE
# - Can set branch to specific commit to ensure no changes are made without knowledge
#   |-----> changed to commit id which is tied to branch and more stable
# - Compatibility defined to a single engine file
#   |-----> Adapters must be compatible with the given engine
# - Experiment configs are defined in the experiment_configs folder
#   |-----> NOTE: AT LEAST TWO EXPERIMENT CONFIGS MUST BE DEFINED
#       |-----> This is so that it triggers the selection swap on the server side
# - OBSERVED STATES MUST MATCH ADAPTER SELECTION NAMING
#    |-----> ENCODED STATES MUST MATCH OBSERVED STATES NAMING
#    |-----> For multiple versions from same adapter, use -v2, -v3, etc.
class Applications:
    def __init__(self):
        self.data ={
            "Sailing":{
                "github_user": "pdfosborne",
                "repository": "elsciRL-App-Sailing",
                "commit_id": "*",
                "engine_folder": "environment",
                "engine_filename": "engine",
                "config_folder": "configs",
                "experiment_config_filenames": {"quick-test":"testing.json", 
                                                "Osborne-2024":"config.json"},
                "local_config_filenames": {"easy":"config_local.json"},
                "local_adapter_folder": "adapters",
                "adapter_filenames": {"default":"default", "language":"language", "LLM":"LLM_adapter"},
                "local_analysis_folder": "analysis",
                "local_analysis_filenames": {"sailing_graphs":"sailing_graphs"},
                "prerender_data_folder": "prerender",
                "prerender_data_filenames": {"language-v1":"observed_states_language_1000000.txt",
                                            "LLM-llama-v1":"observed_states_Sailing_easy_LLM_llama32_100000_v1.jsonl"},
                "prerender_data_encoded_filenames":{},
                "prerender_image_filenames": {"Setup":"sailing_setup.png"},
                "instruction_folder": "instructions",
                "instruction_filenames": {"osborne2025":"Osborne2025_instruction_results_Sailing_LLM-llama-v1.json"}
                },
            "Classroom":{
                "github_user": "pdfosborne",
                "repository": "elsciRL-App-Classroom",
                "commit_id": "*",
                "engine_folder": "environment",
                "engine_filename": "engine",
                "config_folder": "configs",
                "experiment_config_filenames": {"default":"config.json"},
                "local_config_filenames": {"classroom_A":"classroom_A.json"},
                "local_adapter_folder": "adapters",
                "adapter_filenames": {"default":"default", "classroom_A_language":"classroom_A_language", "LLM":"LLM_adapter"},
                "local_analysis_folder": "analysis",
                "local_analysis_filenames": {},
                "prerender_data_folder": "prerender",
                "prerender_data_filenames": {"classroom_A_language-v1":"observed_states_Classroom_classroom_A_classroom_A_language_100.txt", 
                                            "LLM-v1":"observed_states_Classroom_classroom_A_LLM_llama32_100.txt"},
                "prerender_data_encoded_filenames":{"classroom_A_language-v1":"encoded_observed_states_Classroom_classroom_A_classroom_A_language_100.txt",
                                                    "LLM-v1":"encoded_observed_states_Classroom_classroom_A_LLM_llama32_100.txt"},
                "prerender_image_filenames": {"Classroom_A_Setup":"Classroom_A_Summary.png"},
                "instruction_folder": "instructions",
                "instruction_filenames": {"test":"instruction_results_Classroom_test.json",
                                          "osborne2025":"Osborne2025_instruction_results_Classroom_LLM.json"}
                },
            "Gym-FrozenLake":{
                "github_user": "pdfosborne",
                "repository": "elsciRL-App-GymFrozenLake",
                "commit_id": "*",
                "engine_folder": "environment",
                "engine_filename": "engine",
                "config_folder": "configs",
                "experiment_config_filenames": {"quick_test":"fast_agent.json", "Osborne2024_agent":"Osborne2024_agent.json"},
                "local_config_filenames": {"Osborne2024_env":"Osborne2024_env.json"},
                "local_adapter_folder": "adapters",
                "adapter_filenames": {"numeric_encoder":"numeric", "language":"language", "LLM":"LLM_adapter"},
                "local_analysis_folder": "analysis",
                "local_analysis_filenames": {},
                "prerender_data_folder": "prerender",
                "prerender_data_filenames": {"language-v1":"observed_states_Gym-FrozenLake_Osborne2024_env_language_100.txt", 
                                            "LLM-v1":"observed_states_Gym-FrozenLake_Osborne2024_env_LLM_llama32_100.txt"},
                "prerender_data_encoded_filenames": {"language-v1":"encoded_observed_states_Gym-FrozenLake_Osborne2024_env_language_100.txt",
                                            "LLM-v1":"encoded_observed_states_Gym-FrozenLake_Osborne2024_env_LLM_llama32_100.txt"},
                "prerender_image_filenames": {"FrozenLake_Setup":"FrozenLake_4x4.png"},
                "instruction_folder": "instructions",
                "instruction_filenames": {"osborne2025":"Osborne2025_instruction_results_Gym-FrozenLake_LLM.json"}
                },
            "Chess":{
                "github_user": "pdfosborne",
                "repository": "elsciRL-App-Chess",
                "commit_id": "*",
                "engine_folder": "environment",
                "engine_filename": "engine",
                "config_folder": "configs",
                "experiment_config_filenames": {"Osborne2024_agent":"config.json"},
                "local_config_filenames": {"Osborne2024_env":"config_local.json"},
                "local_adapter_folder": "adapters",
                "adapter_filenames": {"numeric_board_mapping": "numeric_board",
                                      "numeric_piece_counter":"numeric_piece_counter", 
                                      "active_pieces_language":"language_active_pieces",
                                      "LLM":"LLM_adapter"},
                "local_analysis_folder": "analysis",
                "local_analysis_filenames": {},
                "prerender_data_folder": "prerender",
                "prerender_data_filenames": {"active_pieces_language-v1":"observed_states_active_pieces_language_50000.txt",
                                             "LLM-v1": "observed_states_Chess_Osborne2024_env_LLM_llama32_100000.jsonl",},
                "prerender_data_encoded_filenames":{},
                "prerender_image_filenames": {"Board_Setup":"board_start.png"},
                "instruction_folder": "instructions",
                "instruction_filenames": {"osborne2025":"Osborne2025_instruction_results_Chess_LLM-v1.json"}
                },
            # "TextWorldExpress":{
            #     "github_user": "pdfosborne",
            #     "repository": "elsciRL-App-TextWorldExpress",
            #     "commit_id": "*",
            #     "engine_folder": "environment",
            #     "engine_filename": "engine",
            #     "config_folder": "configs",
            #     "experiment_config_filenames": {"Osborne2024_agent":"config.json"},
            #     "local_config_filenames": {"twc-easy":"twc-easy.json", 
            #                                 "twc-medium":"twc-medium.json",
            #                                 "twc-hard":"twc-hard.json", 
            #                                 "cookingworld-easy":"cookingworld-easy.json",
            #                                 "cookingworld-medium":"cookingworld-medium.json",
            #                                 "cookingworld-hard":"cookingworld-hard.json", 
            #                                 "coincollector":"coincollector.json",
            #                                 "mapreader":"mapreader.json"},
            #     "local_adapter_folder": "adapters",
            #     "adapter_filenames": {"language_default":"language", "LLM":"LLM_adapter"},
            #     "local_analysis_folder": "analysis",
            #     "local_analysis_filenames": {},
            #     "prerender_data_folder": "prerender",
            #     "prerender_data_filenames": {
            #         "language-v1":
            #             "observed_states_TextWorldExpress_twc-easy_language_default_10000_29-05-2025_15-53.txt",
            #         "LLM-v1":
            #             "observed_states_TextWorldExpress_cookingworld-easy_language_default_10000_29-05-2025_16-02.txt",},
            #     "prerender_data_encoded_filenames":{},
            #     "prerender_image_filenames": {},
            #     "instruction_folder": "instructions",
            #     "instruction_filenames": {}
            #     },
            "Maze":{
                "github_user": "pdfosborne",
                "repository": "elsciRL-App-Maze",
                "commit_id": "*",
                "engine_folder": "environment",
                "engine_filename": "engine",
                "config_folder": "configs",
                "experiment_config_filenames": {"default_agent":"config.json"},
                "local_config_filenames": {"umaze":"umaze.json",
                                           "double_t_maze":"double_t_maze.json",
                                           "medium":"medium.json",
                                           "large":"large.json",
                                           #"random":"random.json"
                                           },
                "local_adapter_folder": "adapters",
                "adapter_filenames": {"language_default":"language", "LLM":"LLM_adapter"},
                "local_analysis_folder": "analysis",
                "local_analysis_filenames": {},
                "prerender_data_folder": "prerender",
                "prerender_data_filenames": {'language-umaze-v1':'observed_states_Maze_umaze_language_default_1000.txt',
                                            'LLM-umaze-v1':'observed_states_Maze_umaze_LLM_llama32_100.txt',
                                            'language-double_t_maze-v1':'observed_states_Maze_double_t_maze_language_default_1000.txt',
                                            'LLM-double_t_maze-v1':'observed_states_Maze_double_t_maze_LLM_llama32_1000.txt',
                                            'language-medium-v1':'observed_states_Maze_medium_language_default_1000.txt',
                                            'LLM-medium-v1':'observed_states_Maze_medium_LLM_llama32_1000.txt',
                                            'language-large-v1':'observed_states_Maze_large_language_default_1000.txt',
                                            'LLM-large-v1':'observed_states_Maze_large_LLM_llama32_1000.txt'
                                            },
                "prerender_data_encoded_filenames":{'language-umaze-v1':'encoded_observed_states_Maze_umaze_language_default_1000.txt',
                                            'LLM-umaze-v1':'encoded_observed_states_Maze_umaze_LLM_llama32_100.txt',
                                            'language-double_t_maze-v1':'encoded_observed_states_Maze_double_t_maze_language_default_1000.txt',
                                            'LLM-double_t_maze-v1':'encoded_observed_states_Maze_double_t_maze_LLM_llama32_1000.txt',
                                            'language-medium-v1':'encoded_observed_states_Maze_medium_language_default_1000.txt',
                                            'LLM-medium-v1':'encoded_observed_states_Maze_medium_LLM_llama32_1000.txt',
                                            'language-large-v1':'encoded_observed_states_Maze_large_language_default_1000.txt',
                                            'LLM-large-v1':'encoded_observed_states_Maze_large_LLM_llama32_1000.txt'
                                            },
                "prerender_image_filenames": {},
                "instruction_folder": "instructions",
                "instruction_filenames": {"osborne2025umaze":"Osborne2025_instruction_results_Maze_LLM-umaze-v1.json",
                                          "osborne2025double_t_maze":"Osborne2025_instruction_results_Maze_LLM-double_t_maze-v1.json",
                                          "osborne2025medium":"Osborne2025_instruction_results_Maze_LLM-medium-v1.json",
                                          "osborne2025large":"Osborne2025_instruction_results_Maze_LLM-large-v1.json"}
                }
        }


# Get application data
import numpy as np
from elsciRL.application_suite.import_data import Applications
from elsciRL.application_suite.import_tool import PullApplications

# Get search method
from elsciRL.instruction_following.elsciRL_GUI_search import elsciRLSearch as elsci_search
import os
import json
import torch
from torch import Tensor
from datetime import datetime
import ollama
from tqdm import tqdm

# Get language encoder
from elsciRL.encoders.language_transformers.MiniLM_L6v2 import LanguageEncoder as MiniLM_L6v2

class Prerender:
    def __init__(self) -> None:
         # Get application data
        imports = Applications().data
        self.possible_applications = list(imports.keys())
        pull_data = PullApplications()
        self.pull_app_data = pull_data.pull(problem_selection=self.possible_applications)
        self.ExperimentConfig = pull_data.setup()
        self.observed_states = None


    def get_observed_states(self, engine, 
            selected_application:str='', selected_config:str='', selected_adapter:list=[''],
            local_config:dict={}, 
            adapters:dict={},  
            num_explor_episodes:int=1000):
            
        # Search Save Directory
        if not os.path.exists('./prerender-data'):
            os.mkdir('./prerender-data')
                            
        time = datetime.now().strftime("%d-%m-%Y_%H-%M")
        save_dir = './prerender-data/' #+ str('search') + '_' + time
        if not os.path.exists(save_dir):                
            os.mkdir(save_dir)
        # UPDATE EXPERIMENT CONFIG FOR SEARCH
        # - only use q learn tab agent
        selected_agents =  ['Qlearntab']
        self.ExperimentConfig.update({
            'number_training_episodes': int(num_explor_episodes),
            'agent_select': selected_agents         
        })
        local_config.update({   
            'adapter_select': selected_adapter    
        })

        elsci_run = elsci_search(Config=self.ExperimentConfig,
                                    LocalConfig=local_config,
                                    Engine=engine, Adapters=adapters,
                                    save_dir=save_dir,
                                    number_exploration_episodes=num_explor_episodes,
                                    match_sim_threshold=0.9,
                                    observed_states=None)
        observed_states = elsci_run.search()
        print(f"\nNumber of observed states: {len(observed_states)}")
        # Create save directory structure if not exists
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)

        if not os.path.exists(os.path.join(save_dir, selected_application)):
            os.mkdir(os.path.join(save_dir, selected_application))

        if not os.path.exists(os.path.join(save_dir, selected_application, selected_adapter[0])):
            os.mkdir(os.path.join(save_dir, selected_application, selected_adapter[0]))
        # Add problem name to save_dir
        save_dir = os.path.join(save_dir, selected_application, selected_adapter[0])


        # Update file name to include problem name, number of episodes, and timestamp
        if selected_adapter[0].split('_')[0] == 'LLM':
            # i.e. if the model name is 'llama3.2', we want to shorten it to 'llama32'
            if len({local_config['model_name'].split('.')}) > 2:
                mod_name_short = {local_config['model_name'].split('.')[0]}+{local_config['model_name'].split('.')[1]}
            else:
                mod_name_short = local_config['model_name'].split('.')[0]
            file_name = f"observed_states_{selected_application}_{selected_config}_{selected_adapter[0]}_{mod_name_short}_{num_explor_episodes}_{time}.txt"
        else:
            file_name = f"observed_states_{selected_application}_{selected_config}_{selected_adapter[0]}_{num_explor_episodes}_{time}.txt"
        file_path = os.path.join(save_dir, file_name)

        # Save observed states only
        with open(file_path, 'w') as f:
            json.dump(observed_states, f)
            f.close()
        
        return observed_states, file_name
            


    def run(self):

        # Allow terminal input to select application
        print("Select an application from the following options:")
        for i, app in enumerate(self.possible_applications):
            print(f"{i + 1}: {app}")

        selected_index = int(input("Enter the number of the application you want to select: ")) - 1
        selected_application = self.possible_applications[selected_index]

        # Allow terminal input to select configuration
        print("Select a configuration from the following options:")
        for i, config in enumerate(self.pull_app_data[selected_application]['local_configs']):
            print(f"{i + 1}: {config}")

        config_input_index = int(input("Enter the number of the configuration you want to select: ")) - 1
        config_input = list(self.pull_app_data[selected_application]['local_configs'].keys())[config_input_index]

        # Allow terminal input to select adapter
        adapter_list = self.pull_app_data[selected_application]['adapters']
        print("Select an adapter from the following options:")
        for i, adapter in enumerate(adapter_list):
            print(f"{i + 1}: {adapter}")

        selected_adapter_index = int(input("Enter the number of the adapter you want to select: ")) - 1
        selected_adapter = [list(adapter_list.keys())[selected_adapter_index]]

        # Get data for the selected application
        engine = self.pull_app_data[selected_application]['engine']
        local_config = self.pull_app_data[selected_application]['local_configs'][config_input]
        adapters = self.pull_app_data[selected_application]['adapters']

        if selected_adapter[0].split('_')[0] == 'LLM':
            ollama_models = ollama.list()
            print("Select an ollama model from the following options:")
            for i, model in enumerate(ollama_models['models']):
                print(f"{i + 1}: {model['model']}")
            select_ollama_model_index = int(input("Enter the number of the ollama model you want to select: ")) - 1
            select_ollama_model = ollama_models['models'][select_ollama_model_index]['model']
            local_config['model_name'] = select_ollama_model
            print(f"-- Selected ollama model: {select_ollama_model}")
        
        num_explor_episodes = int(input("Enter the number of the exploration episodes: "))


        print("--------------------------------")
        print("-Selected options-")
        print(f"-- Selected application: {selected_application}")
        print(f"-- Selected configuration: {config_input}")
        print(f"-- Selected adapter: {selected_adapter[0]}")
        if selected_adapter[0].split('_')[0] == 'LLM':
            print(f"-- Selected ollama model: {select_ollama_model}")
        print(f"-- Number of exploration episodes: {num_explor_episodes}")
        print("--------------------------------")

        self.observed_states, self.observed_states_file_name = self.get_observed_states(engine, 
                                selected_application, config_input, selected_adapter,
                                local_config, 
                                adapters, 
                                num_explor_episodes)
        
    def adapter_apply(self):
        file_names = [file for file in os.listdir('./') if file.endswith('.txt')]
        for n, file in enumerate(file_names):
            print(f"- {n}: {file}")
        selection = input("\n Select the file to transform with the adapter (by number): ")
        selected_file = file_names[int(selection)]

        # Get application name from file name
        selected_application = selected_file.split('_')[2]
        if selected_application not in self.possible_applications:
            # Allow terminal input to select application
            print("Application name not found, please confirm the application name from the following options:")
            for i, app in enumerate(self.possible_applications):
                print(f"{i + 1}: {app}")

            selected_index = int(input("Enter the number of the application you want to select: ")) - 1
            selected_application = self.possible_applications[selected_index]

        # Allow terminal input to select configuration
        print("Select a configuration from the following options:")
        for i, config in enumerate(self.pull_app_data[selected_application]['local_configs']):
            print(f"{i + 1}: {config}")

        config_input_index = int(input("Enter the number of the configuration you want to select: ")) - 1
        config_input = list(self.pull_app_data[selected_application]['local_configs'].keys())[config_input_index]
        
        # Allow terminal input to select adapter
        adapter_list = self.pull_app_data[selected_application]['adapters']
        print("Select an adapter from the following options:")
        for i, adapter in enumerate(adapter_list):
            print(f"{i + 1}: {adapter}")

        selected_adapter_index = int(input("Enter the number of the adapter you want to select: ")) - 1
        selected_adapter = list(adapter_list.keys())[selected_adapter_index]

        # Get data for the selected application
        engine = self.pull_app_data[selected_application]['engine']
        local_config = self.pull_app_data[selected_application]['local_configs'][config_input]
        adapters = self.pull_app_data[selected_application]['adapters']

        if selected_adapter.split('_')[0] == 'LLM':
            ollama_models = ollama.list()
            print("Select an ollama model from the following options:")
            for i, model in enumerate(ollama_models['models']):
                print(f"{i + 1}: {model['model']}")
            select_ollama_model_index = int(input("Enter the number of the ollama model you want to select: ")) - 1
            select_ollama_model = ollama_models['models'][select_ollama_model_index]['model']
            local_config['model_name'] = select_ollama_model
            print(f"-- Selected ollama model: {select_ollama_model}")

        with open(selected_file, 'r') as f:
            observed_data_in = json.load(f)
        number_episodes = len(observed_data_in)
        # Save the transformed data
        time = datetime.now().strftime("%d-%m-%Y_%H-%M")
        print("--------------------------------")
        print("-Selected options-")
        print(f"-- Selected application: {selected_application}")
        print(f"-- Selected adapter: {selected_adapter}")
        if selected_adapter.split('_')[0] == 'LLM':
            print(f"-- Selected ollama model: {select_ollama_model}")
            file_name = 'observed_states_'+selected_application+'_'+config_input+'_'+selected_adapter+'_'+select_ollama_model+'_'+str(number_episodes)+'_'+time+'.txt'
        else:
            file_name = 'observed_states_'+selected_application+'_'+config_input+'_'+selected_adapter+'_'+str(number_episodes)+'_'+time+'.txt'
        print("--------------------------------")

        ENGINE_APPLY = engine(local_setup_info=local_config)
        ADAPTER_APPLY = adapters[selected_adapter](setup_info=local_config)
        print(f"Using engine: {ENGINE_APPLY} to generate legal moves.")
        print(f"Using adapter: {ADAPTER_APPLY}")

        print(f"\n Transforming {len(observed_data_in)} observed states from {selected_file}...\n")
        observed_data_out = {}
        for key in tqdm(observed_data_in.keys(), desc="Transforming observed states with adapter", total=len(observed_data_in)):
            legal_moves_sampled = ENGINE_APPLY.legal_move_generator(obs=key)
            observed_data_out[key] = ADAPTER_APPLY.adapter(state=key, legal_moves=legal_moves_sampled, encode=False)

        with open(file_name, 'w') as f:
            json.dump(observed_data_out, f)


    def encode(self, observed_states:dict|str=None,
                        directory_search:bool=False,
                        save_dir:str=None,
                        encoder:str ='MiniLM_L6v2') -> Tensor:
        """    Encodes the observed states using a language encoder.
        Args:
            observed_states (dict or str): The observed states to encode, can be the dictionary or the directory path string.
            directory_search (bool): Whether to search for the observed states in a directory. Defaults to False.
            save_dir (str): The directory where the encoded states will be saved. If None, defaults to './encoded-prerender-data'.
            encoder (str): The name of the encoder to use. Defaults to 'MiniLM_L6v2', options include:
                - 'MiniLM_L6v2': A lightweight language model suitable for encoding text.
                - ~~Other encoders can be added in the future.~~
        Returns:
            Tensor: The encoded representation of the observed states.
        """
        # ------------------------------------------------------------------
        # Define the available encoders
        # Currently only MiniLM_L6v2 is available, but can be extended in the future.
        ENCODERS = {'MiniLM_L6v2': MiniLM_L6v2}
        encoder = ENCODERS[encoder]()
        # ------------------------------------------------------------------
        if observed_states is None:
            if (self.observed_states is not None) and (not directory_search):
                observed_states = self.observed_states
                save_dir  = './prerender-data/'
                observed_states_filename = self.observed_states_file_name
            else:
                print("\n ----------------------------------------------------")
                print(" No observed states provided. Please select a file to encode.")
                print(" ----------------------------------------------------\n")
                file_names = [file for file in os.listdir('./') if file.endswith('.txt')]
                for n, file in enumerate(file_names):
                    print(f"- {n}: {file}")
                selection = input("\n Select the file to encode (by number): ")
                observed_states_filename = file_names[int(selection)]
                observed_states_path = os.path.join('./', observed_states_filename)
                with open(observed_states_path, 'r') as f:
                    observed_states = json.loads(f.read())
                save_dir = './'
        else:
            if isinstance(observed_states, str):
                observed_states_filename = observed_states.split('/')[-1].split('.')[0]
                if not save_dir:
                    save_dir = os.path.dirname(observed_states)
                with open(observed_states, 'r') as f:
                    observed_states = json.loads(f.read())
            else:
                observed_states_filename = 'observed_states'
                if not save_dir:
                    save_dir = './'

        # Encode the observed states
        print(f"\n Encoding observed state file {observed_states_filename} using {encoder.name}...")
        str_states = [str_state for str_state in observed_states.values()]
        observed_states_encoded = encoder.encode(str_states)


        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        file_path = os.path.join(save_dir, 'encoded_' + observed_states_filename.split('.')[0] + '.txt')
        np.savetxt(file_path, observed_states_encoded.numpy())
        print(f"Encoded states saved to {file_path}")
        print(f"Number of States: {len(observed_states_encoded)}")

        return observed_states_encoded


class AgentFactory:
    """Factory for creating agent instances based on type name and parameters."""
    def __init__(self, adapters, setup_info):
        from elsciRL.agents.table_q_agent import TableQLearningAgent
        from elsciRL.agents.DQN import DQNAgent
        from elsciRL.agents.LLM_agents.ollama_agent import LLMAgent as OllamaAgent
        self.adapters = adapters
        self.agent_types = {
            "Qlearntab": TableQLearningAgent,
            "DQN": DQNAgent,
            "LLM_Ollama": OllamaAgent,
        }
        self.setup_info = setup_info

    def register_agent(self, name, agent_class):
        self.agent_types[name] = agent_class

    def create(self, agent_type, agent_parameters, engine=None, adapter=None):
        if agent_type == "DQN":
            if adapter:
                adapter_sample = self.adapters[adapter](setup_info=self.setup_info)
                # Set input_size from adapter
                try:
                    input_size = adapter_sample.input_dim
                    print(f"Using input_dim from adapter {adapter}: {input_size}")
                except Exception:
                    try:
                        input_size = adapter_sample.encoder.output_dim
                        print(f"Using encoder output_dim from encoder {adapter_sample.encoder}: {input_size}")
                    except Exception:
                        try:
                            input_size = adapter_sample.LLM_adapter.encoder.output_dim
                            print(f"Using LLM_adapter encoder output_dim from LLM adapter {adapter_sample.LLM_adapter}: {input_size}")
                        except Exception:
                            print(f"Adapter {adapter} does not have input_dim specified.")
                            raise ValueError(f"No input dim size found in adapter: {adapter}")

            if engine:
                print(engine)
                engine_sample = engine(local_setup_info=self.setup_info)
                try:
                    output_size = engine_sample.output_size
                except Exception:
                    try:
                        output_size = engine_sample.output_dim
                    except Exception:
                        try:
                            output_size = engine_sample.output_dim_size
                        except Exception:
                            print(f"Engine {engine} does not contain output dim size for DQN agent, using default 1,000.")
                            output_size = 1000
            # Order must match DQN input
            temp_dict = {'input_size': input_size, 'output_size': output_size}
            temp_dict.update(agent_parameters)
        else:
            # For other agents, we assume the parameters are already in the correct format
            temp_dict = agent_parameters
        if agent_type not in self.agent_types:
            raise ValueError(f"Unknown agent type: {agent_type}")
        return self.agent_types[agent_type](**temp_dict)



import time
import numpy as np
from tqdm import tqdm

# ------ Imports -----------------------------------------
# Agent Setup
from elsciRL.environment_setup.imports import ImportHelper

# Evaluation standards
from elsciRL.environment_setup.results_table import ResultsTable
from elsciRL.environment_setup.elsciRL_info import elsciRLInfo


def episode_loop(Engine, Adapters: dict, local_setup_info: dict, number_episodes: int = 1000, 
                 batch_number: int = 0, observed_states: dict = {}) -> dict:
    # --- INIT state space from engine
    agent_adapter_name = local_setup_info['agent_type'] + "_" + local_setup_info['adapter_select']
    engine = Engine(local_setup_info)
    start_obs = engine.reset()
    # --- PRESET elsciRL INFO
    # Agent
    Imports = ImportHelper(local_setup_info)
    agent, agent_type, agent_name, agent_state_adapter = (
        Imports.agent_info(Adapters)
    )
    (
        num_train_episodes,
        num_test_episodes,
        training_action_cap,
        testing_action_cap,
        reward_signal,
    ) = Imports.parameter_info()

    # Training or testing phase flag
    train = Imports.training_flag()

    # Mode selection (already initialized)
    # --- elsciRL
    live_env, observed_states_flag = (
        Imports.live_env_flag()
    )
    # Results formatting
    results = ResultsTable(local_setup_info)
    # elsciRL input function
    # - We only want to init trackers on first batch otherwise it resets knowledge
    elsciRL = elsciRLInfo(observed_states)
    # RENDER AND SUB-GOALS REMOVED COMPLETELY SO SAVE RUN-TIME
    
    for episode in tqdm(range(0, number_episodes)):
        action_history = []
        # ---
        # Start observation is used instead of .reset()  fn so that this can be overridden for repeat analysis from the same start pos
        obs = engine.reset(start_obs=start_obs)
        legal_moves = engine.legal_move_generator(obs)

        # LLM agents need to pass the state as a string
        if agent_type.split("_")[0] == "LLM":
            state = agent_state_adapter.adapter(
            state=obs,
            legal_moves=legal_moves,
            episode_action_history=action_history,
            encode=False,
        )
        else:
            state = agent_state_adapter.adapter(
                state=obs,
                legal_moves=legal_moves,
                episode_action_history=action_history,
                encode=True,
            )
        # ---
        start_time = time.time()
        episode_reward: int = 0
        # ---
        for action in range(0, training_action_cap):
            if live_env:
                # Agent takes action
                legal_moves = engine.legal_move_generator(obs)
                agent_action = agent.policy(state, legal_moves)

                if isinstance(agent_action, np.int64):
                    action_history.append(agent_action.item())
                else:
                    action_history.append(agent_action)

                next_obs, reward, terminated, _ = engine.step(
                    state=obs, action=agent_action
                )
                
                # Can override reward per action with small negative punishment
                if reward == 0:
                    reward = reward_signal[1]

                # Only update observed states if not already observed
                if next_obs not in observed_states:
                    legal_moves = engine.legal_move_generator(next_obs)
                    # LLM agents need to pass the state as a string
                    if agent_type.split("_")[0] == "LLM":
                        next_state = agent_state_adapter.adapter(
                        state=next_obs,
                        legal_moves=legal_moves,
                        episode_action_history=action_history,
                        encode=False,
                    )
                    else:
                        next_state = agent_state_adapter.adapter(
                            state=next_obs,
                            legal_moves=legal_moves,
                            episode_action_history=action_history,
                            encode=True,
                        )
                    # elsciRL trackers
                    # TODO: Consider adding prior action history to the tracker so that we can 
                    #  transform observed data across adapters without loss of information
                    observed_states = elsciRL.observed_state_tracker(
                        engine_observation=next_obs,
                        language_state=agent_state_adapter.adapter(
                            state=next_obs,
                            legal_moves=legal_moves,
                            episode_action_history=action_history,
                            encode=False,
                        ),
                    )

            episode_reward += reward
            if terminated:
                break
            else:
                state = next_state
                if live_env:
                    obs = next_obs

        # If action limit reached
        if not terminated:
            reward = reward_signal[2]

        end_time = time.time()
        try:
            agent_results = agent.q_result()
        except:
            agent_results = [0, 0]

        if live_env:
            results.results_per_episode(
                agent_name,
                None,
                episode,
                action,
                episode_reward,
                (end_time - start_time),
                action_history,
                agent_results[0],
                agent_results[1],
            )
    # Output GIF image of all episode frames
    return observed_states



import ollama
from typing import Optional, Dict, Any
import json
import logging

class LLMInstructionValidator:
    """
    A class for validating if a given text matches or completes an instruction
    using Large Language Model reasoning via Ollama.
    """
    
    def __init__(self, 
                 model: str = "llama3.2"):
        """
        Initialize the LLM Instruction Validator.
        
        Args:
            model: Ollama model to use for validation (e.g., "llama3.2", "mistral", "codellama").
            temperature: Temperature for LLM responses (lower = more deterministic).
            host: Ollama host URL. If None, uses default localhost.
        """
        self.model = model
        self.logger = logging.getLogger(__name__)
    
    def validate_instruction_completion(self, 
                                      instruction_description: str, 
                                      best_match: str) -> Dict[str, Any]:
        """
        Compare instruction description with best match to determine if the 
        best match completes or fulfills the instruction.
        
        Args:
            instruction_description: The original instruction or task description
            best_match: The text/response that potentially completes the instruction
            
        Returns:
            Dict containing:
                - 'is_complete': Boolean indicating if instruction is completed
                - 'confidence': Float between 0-1 indicating confidence level
                - 'reasoning': String explaining the LLM's reasoning
                - 'partial_completion': Boolean if partially completed
        """
        
        # Construct the prompt for the LLM
        prompt = self._construct_validation_prompt(instruction_description, best_match)
        
        try:
            response = ollama.chat(
                model=self.model,
                messages=[
                    {
                        "role": "system",
                        "content": "You are an expert at evaluating whether responses complete given instructions. "
                                 "You must respond with valid JSON format."
                    },
                    {
                        "role": "user", 
                        "content": prompt
                    }
                ],
                options={
                    "num_predict": 500
                }
            )
            
            # Parse the LLM response
            result = self._parse_llm_response(response['message']['content'])
            
            self.logger.info(f"Validation completed. Is complete: {result['is_complete']}")
            return result
            
        except Exception as e:
            self.logger.error(f"Error during LLM validation: {str(e)}")
            return {
                'is_complete': False,
                'confidence': 0.0,
                'reasoning': f"Error occurred during validation: {str(e)}",
                'partial_completion': False
            }
    
    def _construct_validation_prompt(self, instruction: str, match: str) -> str:
        """
        Construct the prompt for the LLM to evaluate instruction completion.
        
        Args:
            instruction: The instruction description
            match: The best match text
            
        Returns:
            Formatted prompt string
        """
        prompt = f"""
Please evaluate whether the "Best Match" text completes or fulfills the given "Instruction".
You only need to confirm that the language match well and do not need to check if the best match would update the environment.
The language structure from the environment is fixed and does not change, so do not expect a 'Best Match' that is better structured than what is given.
You need to determine if the current 'Best Match' is likely to be the best match for the instruction given the language structure of the environment, do not expect more detail than what is given.

INSTRUCTION:
{instruction}

BEST MATCH:
{match}

Analyze if the Best Match adequately completes, addresses, or fulfills the Instruction. Consider:
1. Does it directly address what was asked?
2. Is the response complete and comprehensive?
3. Does it meet the intent of the instruction?

Respond ONLY with valid JSON in this exact format:
{{
    "is_complete": true/false,
    "confidence": 0.0-1.0,
    "reasoning": "Brief explanation of your evaluation",
    "partial_completion": true/false
}}

Your confidence should be:
- 0.9-1.0: Very confident the instruction is completed
- 0.7-0.8: Mostly confident but some minor gaps
- 0.5-0.6: Partially completed with significant gaps
- 0.0-0.4: Does not complete the instruction
"""
        return prompt
    
    def _parse_llm_response(self, response_text: str) -> Dict[str, Any]:
        """
        Parse the LLM response and extract validation results.
        
        Args:
            response_text: Raw text response from LLM
            
        Returns:
            Parsed validation results
        """
        try:
            # Clean and parse JSON response
            cleaned_response = response_text.strip()
            
            # Handle potential markdown code blocks
            if "```json" in cleaned_response:
                start = cleaned_response.find("```json") + 7
                end = cleaned_response.find("```", start)
                cleaned_response = cleaned_response[start:end].strip()
            elif "```" in cleaned_response:
                start = cleaned_response.find("```") + 3
                end = cleaned_response.find("```", start)
                cleaned_response = cleaned_response[start:end].strip()
            
            result = json.loads(cleaned_response)
            
            # Validate required fields
            required_fields = ['is_complete', 'confidence', 'reasoning', 'partial_completion']
            for field in required_fields:
                if field not in result:
                    raise ValueError(f"Missing required field: {field}")
            
            # Ensure confidence is between 0 and 1
            result['confidence'] = max(0.0, min(1.0, float(result['confidence'])))
            
            return result
            
        except (json.JSONDecodeError, ValueError, KeyError) as e:
            self.logger.error(f"Failed to parse LLM response: {str(e)}")
            return {
                'is_complete': False,
                'confidence': 0.0,
                'reasoning': f"Failed to parse LLM response: {response_text[:100]}...",
                'partial_completion': False
            }
    
    def batch_validate(self, instruction_match_pairs: list) -> list:
        """
        Validate multiple instruction-match pairs in batch.
        
        Args:
            instruction_match_pairs: List of tuples (instruction, best_match)
            
        Returns:
            List of validation results
        """
        results = []
        for instruction, match in instruction_match_pairs:
            result = self.validate_instruction_completion(instruction, match)
            results.append(result)
        
        return results
    
    def list_available_models(self) -> list:
        """
        List available Ollama models.
        
        Returns:
            List of available model names
        """
        try:
            models = ollama.list()
            return [model['model'].split(':')[0] for model in models['models']]
        except Exception as e:
            self.logger.error(f"Error listing models: {str(e)}")
            return []


# Example usage and convenience function
def validate_instruction_match(instruction_description: str, 
                             best_match: str,
                             model: str = "llama3.2") -> Dict[str, Any]:
    """
    Convenience function to quickly validate if a best match completes an instruction.
    
    Args:
        instruction_description: The instruction to validate against
        best_match: The text that potentially completes the instruction
        model: Ollama model to use (default: "llama2")
        host: Ollama host URL (optional)
        
    Returns:
        Validation result dictionary
    """
    validator = LLMInstructionValidator(model=model)
    return validator.validate_instruction_completion(instruction_description, best_match)



import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np
from collections import deque
import random
import pickle
from typing import List, Tuple, Dict, Any, Hashable, Iterable

from elsciRL.agents.agent_abstract import QLearningAgent

class DQN(nn.Module):
    def __init__(self, input_size: int, output_size: int, hidden_size: int = 128):
        super(DQN, self).__init__()
        self.network = nn.Sequential(
            nn.Linear(input_size, hidden_size),
            nn.ReLU(),
            nn.Linear(hidden_size, hidden_size),
            nn.ReLU(),
            nn.Linear(hidden_size, output_size)
        )
    
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.network(x)

class DQNAgent(QLearningAgent):
    def __init__(self, 
                 input_size: int,
                 output_size: int,
                 hidden_size: int = None,
                 learning_rate: float = 0.001,
                 gamma: float = 0.99,
                 epsilon: float = 1.0,
                 epsilon_min: float = 0.01,
                 epsilon_decay: float = 0.995,
                 memory_size: int = 10000,
                 batch_size: int = 64,
                 target_update: int = 10,
                 ):
        
        self.input_size = input_size
        self.output_size = output_size
        
        
        self.memory = deque(maxlen=memory_size)
        self.batch_size = batch_size
        self.gamma = gamma
        self.epsilon_reset = epsilon
        self.epsilon = epsilon
        self.epsilon_min = epsilon_min
        self.epsilon_decay = epsilon_decay
        self.target_update = target_update
        self.update_counter = 0
        
        # Create main and target networks
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.policy_net = DQN(input_size, output_size, hidden_size).to(self.device)
        print(self.policy_net)
        self.target_net = DQN(input_size, output_size, hidden_size).to(self.device)
        self.target_net.load_state_dict(self.policy_net.state_dict())
        
        self.optimizer = optim.Adam(self.policy_net.parameters(), lr=learning_rate)
        self.criterion = nn.MSELoss()

    def save(self) -> List[Dict]:
        """Save the agent's state"""
        return [
            self.policy_net.state_dict(),
            self.target_net.state_dict(),
            self.optimizer.state_dict(),
            {
                'epsilon': self.epsilon,
                'update_counter': self.update_counter,
                'memory': list(self.memory)  # Convert deque to list for serialization
            }
        ]
    
    def load(self, saved_agent: List[Dict]):
        """Load the agent's state"""
        if len(saved_agent) >= 4:
            self.policy_net.load_state_dict(saved_agent[0])
            self.target_net.load_state_dict(saved_agent[1])
            self.optimizer.load_state_dict(saved_agent[2])
            state_dict = saved_agent[3]
            self.epsilon = state_dict['epsilon']
            self.update_counter = state_dict['update_counter']
            if 'memory' in state_dict:
                self.memory = deque(state_dict['memory'], maxlen=self.memory.maxlen)

    def exploration_parameter_reset(self):
        """Reset the exploration parameter to its initial value"""
        self.epsilon = self.epsilon_reset

    def clone(self):
        """Create a deep copy of the agent"""
        clone = pickle.loads(pickle.dumps(self))
        clone.epsilon = self.epsilon_reset
        return clone

    def policy(self, state: torch.Tensor, legal_actions: list, **kwargs) -> Hashable:
        """Select action using epsilon-greedy policy"""
        if random.random() < self.epsilon:
            # Decay epsilon
            if self.epsilon > self.epsilon_min:
                self.epsilon *= self.epsilon_decay
            return random.choice(legal_actions)
        
        with torch.no_grad():
            state = state.to(self.device)
            # Ensure state has correct shape [batch_size, input_size]
            if len(state.shape) == 1:
                state = state.unsqueeze(0)
            q_values = self.policy_net(state)
            
            # Mask illegal actions with large negative values
            mask = torch.ones_like(q_values) * float('-inf')
            for action in legal_actions:
                mask[0][action] = 0
            q_values = q_values + mask
            
            return q_values.argmax().item()

    def learn(self, state: torch.Tensor, next_state: torch.Tensor, 
              immediate_reward: float, action: Hashable, **kwargs) -> None:
        """Store experience and train the network"""
        # Ensure states have correct shape before storing
        if len(state.shape) == 1:
            state = state.unsqueeze(0)
        if isinstance(next_state, torch.Tensor) and len(next_state.shape) == 1:
            next_state = next_state.unsqueeze(0)
            
        # Store experience in replay memory
        self.memory.append((state, action, next_state, immediate_reward))
        
        # Train if enough samples
        if len(self.memory) >= self.batch_size:
            self._train()
            
        # Update target network periodically
        self.update_counter += 1
        if self.update_counter % self.target_update == 0:
            self.target_net.load_state_dict(self.policy_net.state_dict())
            

    def _train(self):
        """Train the network using experience replay"""
        batch = random.sample(self.memory, self.batch_size)
        states, actions, next_states, rewards = zip(*batch)
        
        # Convert to tensors and ensure correct shapes
        states = torch.cat(states).to(self.device)  # [batch_size, input_size]
        next_states = torch.cat(next_states).to(self.device)  # [batch_size, input_size]
        actions = torch.tensor(actions, device=self.device).long()  # [batch_size]
        rewards = torch.tensor(rewards, device=self.device).float()  # [batch_size]
        
        # Get current Q values
        current_q_values = self.policy_net(states)  # [batch_size, output_size]
        current_q_values = current_q_values.gather(1, actions.unsqueeze(1)).squeeze(1)  # [batch_size]
        
        # Get next Q values from target network
        with torch.no_grad():
            next_q_values = self.target_net(next_states).max(1)[0][:self.batch_size]  # [batch_size]
        
        # Compute target Q values
        target_q_values = rewards + (self.gamma * next_q_values)
        
        # Compute loss and update
        loss = self.criterion(current_q_values, target_q_values)
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step() 


import os
import json

def ensure_dir(path):
    if not os.path.exists(path):
        os.makedirs(path)


def load_config(config_path):
    with open(config_path, 'r') as f:
        if config_path.endswith('.json'):
            return json.load(f)
        # Add more config formats if needed
        raise ValueError("Unsupported config file format.")


def merge_configs(config1, config2):
    # Simple dict merge, can be improved for deep merge
    merged = config1.copy()
    merged.update(config2)
    return merged



import torch
import numpy as np
from torch import Tensor
from elsciRL.encoders.language_transformers.MiniLM_L6v2 import LanguageEncoder as MiniLM_L6v2

# Get search method
import os
import json
from datetime import datetime


def encode_prerender_data(observed_states:dict|str=None,
                        save_dir:str=None,
                        encoder:str ='MiniLM_L6v2') -> Tensor:
    """    Encodes the observed states using a language encoder.
    Args:
        observed_states (dict or str): The observed states to encode, can be the dictionary or the directory path string.
        save_dir (str): The directory where the encoded states will be saved. If None, defaults to './encoded-prerender-data'.
        encoder (str): The name of the encoder to use. Defaults to 'MiniLM_L6v2', options include:
            - 'MiniLM_L6v2': A lightweight language model suitable for encoding text.
            - ~~Other encoders can be added in the future.~~
    Returns:
        Tensor: The encoded representation of the observed states.
    """
    # ------------------------------------------------------------------
    # Define the available encoders
    # Currently only MiniLM_L6v2 is available, but can be extended in the future.
    ENCODERS = {'MiniLM_L6v2': MiniLM_L6v2}
    encoder = ENCODERS[encoder]()
    # ------------------------------------------------------------------
    if observed_states is None:
        print("\n ----------------------------------------------------")
        print(" No observed states provided. Please select a file to encode.")
        print(" ----------------------------------------------------\n")
        file_names = [file for file in os.listdir('./') if file.endswith('.txt')]
        for n, file in enumerate(file_names):
            print(f"- {n}: {file}")
        selection = input("\n Select the file to encode (by number): ")
        observed_states_filename = file_names[int(selection)]
        observed_states_path = os.path.join('./', observed_states_filename)
        with open(observed_states_path, 'r') as f:
            observed_states = json.loads(f.read())
        save_dir = './'
    else:
        if isinstance(observed_states, str):
            observed_states_filename = observed_states.split('/')[-1].split('.')[0]
            if not save_dir:
                save_dir = os.path.dirname(observed_states)
            with open(observed_states, 'r') as f:
                observed_states = json.loads(f.read())
        else:
            observed_states_filename = 'observed_states'
            if not save_dir:
                save_dir = './'

    # Encode the observed states
    print(f"\n Encoding observed state file {observed_states_filename} using {encoder.name}...")
    str_states = [str_state for str_state in observed_states.values()]
    observed_states_encoded = encoder.encode(str_states)

    if not os.path.exists(save_dir):
        os.makedirs(save_dir)
    file_path = os.path.join(save_dir, 'encoded_' + observed_states_filename.split('.')[0] + '.txt')
    np.savetxt(file_path, observed_states_encoded.numpy())
    print(f"Encoded states saved to {file_path}")
    print(f"Number of States: {len(observed_states_encoded)}")

    return observed_states_encoded


from abc import ABC, abstractmethod
from typing import List, Any
import os
import numpy as np
from torch import Tensor

from elsciRL.adapters.LLM_state_generators.base_prompt import elsciRL_base_prompt

from elsciRL.encoders.language_transformers.MiniLM_L6v2 import LanguageEncoder as MiniLM_L6v2


try:
    import ollama
except ImportError:
    print("Warning: ollama library not found. Please install it with: pip install ollama")
    ollama = None

ENCODERS = {
    "MiniLM_L6v2": MiniLM_L6v2
}

class LLMAdapter(ABC):
    """Convert a general prompt and raw text state into a description of the state."""
    def __init__(self, base_prompt:str=None):
        super().__init__()
        # Define the fields that describe the state features:
        if base_prompt:
            self.base_prompt = elsciRL_base_prompt + base_prompt
        else:
            self.base_prompt = elsciRL_base_prompt

    @abstractmethod
    def _read(raw_state) -> list:
        # Read the data.
        # fill in the feature fields
        raise NotImplementedError


class OllamaAdapter(LLMAdapter):
    """Adapter for local Ollama LLM models."""
    
    def __init__(self, base_prompt:str=None, model_name:str="llama3.2", 
                 context_length:int = 1000,
                 action_history_length:int=10, encoder:str='MiniLM_L6v2'):
        print("Using OllamaAdapter with model:", model_name)
        super().__init__(base_prompt)
        self.model_name = model_name
        self.manual_context_length = context_length

        if ollama is None:
            raise ImportError("ollama library is required. Install it with: pip install ollama")
        
        # Initialize the language encoder for encoding functionality
        self.encoder = ENCODERS[encoder]()
        
        # Set the action history length, lower reduces context size and computation time
        self.action_history_length = action_history_length
        self.state_history = []
        self.prior_action_count = -1  # Track the number of actions taken in the current episode

        self.cache = {}  # Cache for storing previous responses
        self.encoder_cache = {}  # Cache for storing encoded states
        
    
    def _read(self, raw_state) -> list:
        """Read the data and fill in the feature fields."""
        # This method should be implemented based on specific requirements
        # For now, returning the raw state as a list
        return [raw_state] if isinstance(raw_state, str) else raw_state
    
    def call_ollama_api(self, prompt: str):
        """Call the local Ollama API with the given prompt."""
        try:
            response = ollama.chat(
                model=self.model_name,
                messages=[
                    {
                        'role': 'system',
                        'content': self.base_prompt
                    },
                    {
                        'role': 'user',
                        'content': prompt[:self.manual_context_length]
                    }
                ]
            )
            return response
        except Exception as e:
            print(f"Error calling Ollama API: {e}")
            return None

    def process_ollama_response(self, response):
        """Process the response from Ollama API."""
        if response and 'message' in response:
            output_response = response['message']['content']
            return output_response
        return None

    def adapter(self, state: any, legal_moves: list = [], episode_action_history: list = [], encode: bool = True, indexed: bool = False) -> Tensor:
        """Returns the adapted form, may require input flag for encoded or non-encoded output.
        
        Args:
            state: The current state of the environment.
            legal_moves: A list of legal moves for the agent.
            episode_action_history: A list of actions taken in the current episode.
            encode: A flag indicating whether to encode the response.
            indexed: A flag indicating whether to index the response.

        States and actions are typically numeric and so need the LLM to describe them in a way that is useful for the agent.

        In order to convert numeric information to text we track recent states and actions as a means to establish a context for the current state.
        """
        # Check cache for previously processed state
        if str(state) in self.cache:
            processed_response = self.cache[str(state)]
            if encode:
                encoded_response = self.encoder_cache.get(str(state), None)
                if encoded_response is not None:
                    # If we have a cached response but not an encoded one, encode it
                    encoded_response = self.encoder.encode(
                        state=processed_response,
                        legal_actions=legal_moves,
                        episode_action_history=episode_action_history,
                        indexed=indexed
                    )
                    self.encoder_cache[str(state)] = encoded_response
            return encoded_response if encode else processed_response
        # -----------------------------------------------
        else:
            # Build the full context prompt including legal moves and action history
            context_parts = []
            # Add state information
            if state:
                # We can call the adapter multiple times for the same state so we only add the state
                # to history if it is new
                if len(episode_action_history) > self.prior_action_count:
                    self.state_history.append(str(state))
                    self.prior_action_count = len(episode_action_history)
            
                context_parts.append(f"The current state to describe is: {str(state)}")
            # Add legal moves if provided
            # TODO: ADD ACTION MAPPER TO ALL ADAPTERS AS STANDARD FUNCTIONALITY
            if len(legal_moves) > 0:
                context_parts.append(f"Legal moves: {str(legal_moves)}")
            
            # Add action history if provided
            if len(episode_action_history) > 0:
                context_parts.append("The following is a history of the states and actions taken in the current episode.")
                recent_actions = episode_action_history[-self.action_history_length:]  # Last N actions
                for n,prior_action in enumerate(recent_actions):
                    try:
                        context_parts.append(f"Prior state {n}: {str(self.state_history[len(self.state_history)-len(recent_actions)-1+n])}")
                        context_parts.append(f"Action {n}: {str(prior_action)}")
                    except:
                        print("\nError accessing state history or action history.\n")
                        print(episode_action_history)
                        print("----------------------\n")
                        print(recent_actions)
            
            # Combine all context into a single prompt
            full_prompt = " | ".join(context_parts)
            
            # Get LLM response
            adapted_state = self.call_ollama_api(full_prompt)
            processed_response = self.process_ollama_response(adapted_state)
            
            if processed_response is None:
                processed_response = str(state) if state is not None else "No state available"
        
            # Cache the processed response
            if str(state) not in self.cache:
                self.cache[str(state)] = processed_response

            # Handle encoding
            if encode:
                if self.encoder is not None:
                    # Use the LanguageEncoder to encode the response
                    state_encoded = self.encoder.encode(
                        state=processed_response,
                        legal_actions=legal_moves,
                        episode_action_history=episode_action_history,
                        indexed=indexed
                    )
                    # Cache the encoded state
                    if str(state) not in self.encoder_cache:
                        self.encoder_cache[str(state)] = state_encoded
                    return state_encoded
                else:
                    print("Warning: Encoder not available, returning processed response as string")
                    return processed_response
            else:
                return processed_response
        
    def sample(self, state: any=None):
        """Returns a sample of an adapted state form (typically initial position of the environment).
           A generic blackjack exmaple is used if no state is provided."""
        if not state:
            state_player = np.random.randint(0, 11, size=(1, 3))[0]
            if sum(state_player) > 21:
                state_player += ' you are bust'
            state_dealer = np.random.randint(0, 11, size=(1, 1))[0]
            state_blackjack = [state_player, state_dealer]
            state_blackjack = [state_player, state_dealer]

            legal_moves = [0, 1, 2]

            if len(state_player) > 2:
                previous_actions = [0]
            else:
                previous_actions = []

            state = 'You are playing blackjack. State='+str(state_blackjack)+', legal moves='+str(legal_moves)+', previous actions='+str(previous_actions)+'.'

        print("\n --- ADAPTED STATE --- \n")
        print(self.adapter(state, encode=False))
        print("\n --- ADAPTED STATE ENCODED --- \n")
        print(self.adapter(state))


from datetime import datetime
import os
# ====== elsciRL IMPORTS ===============================================
# ------ EXPERIMENT ----------------------------------------------------
from elsciRL.experiments.standard import Experiment as STANDARD_RL
# ------ Visual Analysis -----------------------------------------------
from elsciRL.analysis.combined_variance_visual import combined_variance_analysis_graph as COMBINED_VARIANCE_ANALYSIS_GRAPH
# ----------------------------------------------------------------------

class DemoExperiment:
    def __init__(self):        
        # Create output directory if it doesn't exist
        self.cwd = os.getcwd()+'/elsciRL-EXAMPLE-output'
        if not os.path.exists(self.cwd):
            os.mkdir(self.cwd)
    
    def help(self):
        help_output = """
        This is a demo experiment script for elsciRL.
        It allows you to run a standard RL experiment on a selected problem from the elsciRL application suite.
        The script will guide you through the process of selecting a problem, configuring the experiment, and running it.
        You can also evaluate the results of the experiment after it has been run.
        Usage:
        1. Run the script.
        2. Follow the prompts to select a problem and configure the experiment.
        3. The experiment will be run and the results will be saved in a directory.
        4. You can evaluate the results by calling the evaluate() method.
        Example:
        >>> demo = DemoExperiment()
        >>> demo.run()
        >>> demo.evaluate()
        """
        print(help_output)

    def input(self):
        # ----- User Input -----
        # 1. Number training episodes
        print("Please enter the number of ... (skip to use default) ")
        num_train_epi = input('\t - Training episodes: ')
        if num_train_epi == '':
            num_train_epi = 1000
        else:
            num_train_epi = int(num_train_epi)

        # Update experiment config
        self.num_train_epi = num_train_epi
        # ----------------------

    def results_save_dir(self):
        # Specify save dir
        # - Needs to be performed here in case user changes parameters and re-runs
        time = datetime.now().strftime("%d-%m-%Y_%H-%M")
        self.save_dir = self.cwd+'/test_'+time 
        if not os.path.exists(self.save_dir):
            os.mkdir(self.save_dir)
        # ---

    def experiment(self, problem:str, exp_save_dir:str, num_train_epi:int=0):

        # --- Select local config and experiment config ---
        print("--- LOCAL CONFIGURATION SELECTION ---")
        for i, local_config in enumerate(self.pull_app_data[problem]['local_configs'].keys()):
            print(f"{i+1}. {local_config}")
        local_config_id = input("Please select the local config number (default 1): ")
        if local_config_id.isdigit() and 0 < int(local_config_id) <= len(self.pull_app_data[problem]['local_configs']):
            local_config_id = int(local_config_id) - 1
        else:
            local_config_id = 0
        LocalConfig = self.pull_app_data[problem]['local_configs'][list(self.pull_app_data[problem]['local_configs'].keys())[local_config_id]]
        
        print("\n --- EXPERIMENT CONFIGURATION SELECTION ---")
        for i, experiment_config in enumerate(self.pull_app_data[problem]['experiment_configs'].keys()):
            print(f"{i+1}. {experiment_config}")
        experiment_config_id = input("Please select the experiment config number (default 1): ")
        if experiment_config_id.isdigit() and 0 < int(experiment_config_id) <= len(self.pull_app_data[problem]['experiment_configs']):
            experiment_config_id = int(experiment_config_id) - 1
        else:
            experiment_config_id = 0
        ExperimentConfig = self.pull_app_data[problem]['experiment_configs'][list(self.pull_app_data[problem]['experiment_configs'].keys())[experiment_config_id]]

        if num_train_epi != 0:
            ExperimentConfig['number_training_episodes'] = num_train_epi
            if int(num_train_epi/10) > 10:
                ExperimentConfig['number_test_episodes'] = int(num_train_epi/10)
            else:
                ExperimentConfig['number_test_episodes'] = 10

        # ------------------------------------------------------
        # Adapter Selection
        print("\n --- ADAPTER SELECTION ---")
        for i, adapter in enumerate(self.pull_app_data[problem]['adapters'].keys()):
            if not adapter.startswith('LLM'):
                print(f"{i+1}. {adapter}")
        adapter_id = input("Please select the adapter number (default 1): ")
        if adapter_id.isdigit() and 0 < int(adapter_id) <= len(self.pull_app_data[problem]['adapters']):
            adapter_id = int(adapter_id) - 1
        else:
            adapter_id = 0

        # --------------------------------------------------------------------
        # Set the selected agent
        ExperimentConfig['agent_select'] = ['Qlearntab']
        ExperimentConfig['adapter_select'] = [list(self.pull_app_data[problem]['adapters'].keys())[adapter_id]]
        ExperimentConfig['adapter_input_dict'] = {'Qlearntab': [list(self.pull_app_data[problem]['adapters'].keys())[adapter_id]]}
        if ExperimentConfig['number_training_repeats'] > 1:
            ExperimentConfig['number_training_repeats'] = 5
        if ExperimentConfig['number_training_seeds'] > 1:
            ExperimentConfig['number_training_seeds'] = 5
        # Flat Baselines
        exp = STANDARD_RL(Config=ExperimentConfig, ProblemConfig=LocalConfig, 
                    Engine=self.pull_app_data[problem]['engine'], Adapters=self.pull_app_data[problem]['adapters'],
                    save_dir=exp_save_dir, show_figures = 'No', window_size=0.1)
        # --------------------------------------------------------------------
        return exp

    def run(self):
        # IMPORT HERE SO ITS NOT LOADED ON STARTUP
        from elsciRL.application_suite.import_tool import PullApplications
        self.application_data = PullApplications()
        self.application_list:list=['Classroom', 'Gym-FrozenLake', 'Sailing']
        self.pull_app_data = self.application_data.pull(problem_selection=self.application_list)
        print("\n --- PULLING APPLICATION DATA ---")
        for app in self.application_list:
            print("--------------------------------------------------")
            print(f"Application: {app}")
            print("Engine:", self.pull_app_data[app]['engine'])
            print("Adapters:", self.pull_app_data[app]['adapters'])
            print("Experiment Configs:", self.pull_app_data[app]['experiment_configs'])
            print("Local Configs:", self.pull_app_data[app]['local_configs'])
        print("-------------------------------------------------- \n ")
        # USER INPUTS FOR BASIC SELECTION OPTIONS
        # --- Problem selection ---
        print("\n --- PROBLEM SELECTION ---")
        for i, prob in enumerate(self.application_list):
            print(f"{i+1}. {prob}")
        problem_id = input("Please enter the problem number to run (default 1): ")
        if problem_id.isdigit() and 0 < int(problem_id) <= len(self.application_list):
            problem = self.application_list[int(problem_id) - 1]
        else:
            problem = self.application_list[0]

        # --- TRAINING EPISODES INPUT ---
        num_train_episodes = input("Please enter the number of training episodes (default 1000): ")
        if num_train_episodes == '':
            num_train_episodes = 1000
        try:
            num_train_episodes = int(num_train_episodes)
        except ValueError:
            print("Invalid input for number of training episodes. Using default value of 1000.")
            num_train_episodes = 1000
        # -------------------------------
        
        self.results_save_dir()
        problem_save_dir = self.save_dir + '/' + problem
        if not os.path.exists(problem_save_dir):
            os.mkdir(problem_save_dir)
        print("\n --------------------------------------------------")
        print('Training and Testing on {p} environment'.format(p=problem))
        print("-------------------------------------------------- \n ")
        exp = self.experiment(problem, problem_save_dir, num_train_epi=int(num_train_episodes))
        exp.train()
        exp.test()
        # exp.render_results()

    def evaluate(self):
        COMBINED_VARIANCE_ANALYSIS_GRAPH(self.save_dir, 'TRAINING', show_figures='Yes')
        COMBINED_VARIANCE_ANALYSIS_GRAPH(self.save_dir, 'TESTING', show_figures='Yes')


import random
import numpy as np
import pandas as pd
import json
import pickle
import torch
# LLM libraries
import ollama

import logging
from elsciRL.agents.agent_abstract import LLMAgentAbstract
from elsciRL.encoders.language_transformers.MiniLM_L6v2 import LanguageEncoder

logger = logging.getLogger()
logger.setLevel(logging.INFO)


class LLMAgent(LLMAgentAbstract):
    def __init__(self, epsilon:float=0.2, epsilon_step:float=0.01, model_name: str = "llama3.2", 
                 system_prompt: str = None, context_length: int = 1000, previous_state_action_history_length: int = 10,
                 action_language_mapping: bool = False):
        """
        Initialize the Ollama LLM model for policy-based action selection.
        
        Args:
            model_name (str): Name of the Ollama model to use
            system_prompt (str, optional): System prompt to guide the model's behavior
        """
        # Check if model exists locally, if not pull it
        # List available models
        models = ollama.list()
        model_exists = any(model['model'] == model_name for model in models['models'])

        self.encoder = LanguageEncoder()
        self.cos = torch.nn.CosineSimilarity(dim=0) 
        
        # Set of actions that result in null outcome and so random is taken instead (all lowercase)
        self.null_actions = ['none', 'non', 'null', 'no', 'n', 'na', 'n/a']
        if not model_exists:
            logger.info(f"Model {model_name} not found locally. Please check your model name and try again.")
            print(f"\n ---- \n Available models: {[model['model'] for model in models['models']]}")
        
        # TODO: REMOVED FOR NOW AND INSTEAD JUST CUTTING OF INPUT LENGTH TO SPECIFIED PARAMETER
        # Import all modelfiles from modelfiles directory
        # try:
        #     # TODO: PULLING FROM GITHUB AS LOCAL REFERENCE WOULD REQUIRE USER TO HAVE MODELFILE DEFINED IN THEIR DIRECTORY
        #     # Create llama3.2 model from modelfile
        #     model_name_modelfile = model_name.replace('.', '-')
        #     # ollama.create() doesn't support remote URLs directly
        #     # Need to download the modelfile first
        #     try:
        #         modelfile_url = 'https://raw.githubusercontent.com/pdfosborne/elsciRL/main/elsciRL/agents/LLM_agents/agent_modelfiles/'+model_name_modelfile+'.modelfile'
        #         modelfile_content = urllib.request.urlopen(modelfile_url).read().decode('utf-8')
        #         ollama.create(model=model_name_modelfile, modelfile=modelfile_content)
        #     except Exception as e:
        #         logger.error(f"Error downloading or creating model from modelfile: {e}")
        #     logger.info("Successfully created "+model_name_modelfile+" model from modelfile")
        # except Exception as e:
        #     logger.error(f"Error getting modelfile for model {model_name} from https://github.com/pdfosborne/elsciRL/tree/main/elsciRL/agents/LLM_agents/agent_modelfiles {e}")
        #     logger.info("Using default model instead")
        
        self.manual_context_length = context_length  # Limit context length for LLM to avoid exceeding token limits

        self.model_name = model_name
        self.system_prompt = (
            "You are an AI agent that takes actions based on the current state. "
            "Your task is to analyze the state and select the most appropriate action from the available actions. "
            "Respond with a JSON object containing the selected action and a brief explanation."
        )
        
        # No need to instantiate ollama.Client; use ollama.chat directly
        
        # Store the model for save/load functionality
        self.model = {
            'model_name': model_name,
            'system_prompt': system_prompt
        }
        # Epsilon-greedy exploration parameter
        self.epsilon = epsilon
        self.epsilon_step = epsilon_step
        self.epsilon_reset = epsilon
        self.previous_state_action_history_length = previous_state_action_history_length

        # Diary is used to improve the LLM's decision making based on previous states, actions and rewards
        # TODO: RENAME THE TERM 'DIARY' TO SOMETHING BETTER BECAUSE LLM THINKS ITS WRITING AN ACTUAL DIARY
        self.diary_system_prompt = """
            You are a expected to track the experience of the agent over many episodes.
            You are given a state, action and reward.
            You need to write a summary of the experience based on the state, action and reward that allows the LLM to learn which actions are good and which are bad.
            ONLY select the most appropriate action, do not provide any reasoning, do not provide any explanation, do not provide any additional text, just the action.
            DO NOT RETURN THE ACTION 'THINK' OR 'THINKING' OR 'THINK ABOUT IT' OR 'THINKING ABOUT IT' OR 'THINKING ABOUT THE NEXT ACTION' OR 'THINKING ABOUT THE NEXT MOVE' OR 'THINKING ABOUT THE NEXT STEP' OR 'THINKING ABOUT THE NEXT DECISION' OR 'THINKING ABOUT THE NEXT CHOICE' OR 'THINKING ABOUT THE NEXT OPTION'.

            Respond in JSON format with the following structure:
            {{
                "action": "selected_action",
                "explanation": "brief explanation of the action taken"
            }}
        """
        self.diary = ''
        self.diary_model_name = 'llama3.2'

        # Add system prompt to the LLM prompts if provided
        if system_prompt:
            self.system_prompt = system_prompt + self.system_prompt
            self.diary_system_prompt = system_prompt + self.diary_system_prompt

        # Initial empty value for trajectory history used by LLM to create diary
        self.state_action_history_current = ''

        self.action_language_mapping = action_language_mapping
        self.action_language_mapping_encoder = {}
        self.action_language_mapping_decoder = {}
        self.action_language_mapping_reset = True
        self.action_language_mapping_system_prompt = """
            You are a language mapping agent that maps a single action to a language that the LLM can understand.
            You are given a list of actions and you need to map each action to a language that the LLM can understand.
            The language should be a single phrase that is a common language that the LLM can understand.
            You need to return a JSON object with the following structure:
            {{
                "action": "action",
                "language": "language"
            }}
        """
        self.action_language_mapping_system_prompt += f"Here is some information on the problem: {self.system_prompt}"

    def _LLM_prompt_adjustment(self, state_action_history: str) -> str:
        """
        Adjust the prompt to be more suitable for the LLM based on states, actions and outcomes.
        """
    
        diary = f"""
                Previous log: {self.diary}

                State action history with rewards: {str(state_action_history.split('.')[-self.previous_state_action_history_length:])}

                Please write or update the log entry based on the knowledge obtained from the states, actions and rewards to maximise the reward obtained.
            """
        # Use ollama.chat with the correct message format
        messages = []
        messages.append({'role': 'system', 'content': self.diary_system_prompt})
        messages.append({'role': 'user', 'content': diary})

        response = ollama.chat(
            model=self.diary_model_name,
            messages=messages[:self.manual_context_length]  # Limit context length
        )

        self.diary = response['message']['content']

        return self.diary
    
    def _action_language_mapping(self, state:str, legal_actions:list[str], state_action_history:str) -> str:
        """
        Map the action to a language that the LLM can understand.
        """
        
        # Use previous state action history if available and long enough
        if (state_action_history) and (len(state_action_history.split('.')) >= self.previous_state_action_history_length):
            action_mapping_system_prompt = self.action_language_mapping_system_prompt + f"""Here is the a recent history of states, actions and resulting outcome states, use this to understand the problem:
                {str(state_action_history.split('.')[-self.previous_state_action_history_length:])}"""
            # Reset knowledge after first time it is available
            if self.action_language_mapping_reset:
                print(f"----> Using previous state action history, action language mapping knowledge reset.")
                self.action_language_mapping_reset = False
                self.action_language_mapping_encoder = {}
                self.action_language_mapping_decoder = {}
        else:
            action_mapping_system_prompt = self.action_language_mapping_system_prompt
        # ------------------------------------------------------------
        legal_actions_mapped = []
        for action in legal_actions:
            if action not in self.action_language_mapping_encoder:
                # Convert action to language
                action_mapping = ollama.chat(
                    model=self.model_name,
                    messages=[{'role': 'system', 'content': action_mapping_system_prompt},
                            {'role': 'user', 'content': str(action)}]
                )['message']['content']
                try:
                    # Convert to JSON object
                    action_mapping = json.loads(action_mapping)
                    action_mapping = action_mapping['language']
                except:
                    pass
                action_mapping = action_mapping.split('\n')[-1].strip()  # Take last line of response
                # Use original action in mapping as fallback incase mapping is poor
                action_mapping = 'Original action code was [' + str(action) + '], mapped to language is [' + action_mapping + ']'
                # Store action mapping
                self.action_language_mapping_encoder[action] = action_mapping
                # Decoder used to convert language back to action
                self.action_language_mapping_decoder[action_mapping] = action
                print(f"----> Mapped action {action} to language {action_mapping}")
                legal_actions_mapped.append(action_mapping)
            else:
                # If action already mapped, use existing mapping
                legal_actions_mapped.append(self.action_language_mapping_encoder[action])

        return legal_actions_mapped

    def save(self) -> dict:
        return self.model
    
    def load(self, saved_agent: dict = {}):
        if saved_agent:
            self.model = saved_agent
            self.model_name = saved_agent.get('model_name', 'llama2')
            self.system_prompt = saved_agent.get('system_prompt')
            # No need to re-instantiate a client

    def exploration_parameter_reset(self):
        self.epsilon = self.epsilon_reset
        

    def clone(self):
        clone = pickle.loads(pickle.dumps(self))
        clone.epsilon = self.epsilon_reset
        return clone

    # Fixed order of variables
    def policy(self, state: str, legal_actions: list[str]) -> str:
        """Agent's decision making for next action based on current knowledge and policy type"""

        # Epsilon-greedy action selection to encourage exploration
        if random.random() < self.epsilon:
            action = random.choice(legal_actions)
            logger.info(f"Epsilon-greedy: Random action selected: {action}")   
            if self.epsilon > 0:
                self.epsilon = self.epsilon - (self.epsilon*self.epsilon_step) # Added epsilon step reduction to smooth exploration to greedy
                if self.epsilon < 0:
                    self.epsilon = 0         
        else:
            if self.action_language_mapping:
                print(f"----> Action language mapping enabled, mapping actions to language.")
                legal_actions = self._action_language_mapping(state, legal_actions, self.state_action_history_current)

            prompt = f"""
                        Current state: {state}
                        
                        You must select an action from the following list: {str(legal_actions)}

                        You have access to information on previous states, actions and rewards. Use this information to help you make the best decision: {self.diary}
"""

            # Use ollama.chat with the correct message format
            messages = []
            if self.system_prompt:
                messages.append({'role': 'system', 'content': self.system_prompt})
            messages.append({'role': 'user', 'content': prompt[:self.manual_context_length]})

            if len(prompt) > self.manual_context_length:
                print("\n WARNING: Context length exceeded, cutting off input to fit within model limits.")
                print(" Context Length:", len(prompt))
                print(" Model Limit:", self.manual_context_length)
            response = ollama.chat(
                model=self.model_name,
                messages=messages  # Limit context length
            )
            
            # Result not always ending content with brackets
            # - limit output length to save time, use end of content as action is usually given at the end
            action = response['message']['content'].split('\n')[-1].strip()
            print("\n ----------------------------------------------------")
            # Validate Action against Legal Actions
            if (action in legal_actions) and (action is not None):
                print(f"LLM response content: {action} in LEGAL ACTIONS")
                action = action
            else:
                print(f"Action {action} not in legal actions, trying to find best match.")
                # 1. Check if action is a cross-reference to legal actions
                cross_reference = False
                for term in action.split(' '):
                    if term.strip() in legal_actions:
                        action = term.strip()
                        cross_reference = True
                        break

                if cross_reference is False:
                    # 2. CHECK IF LEGAL ACTION IN ACTION TEXT
                    action_match_found = False
                    for legal_a in legal_actions:
                        if str(legal_a) in action:
                            print(f"Action {action} found in legal moves as {legal_a}.")
                            action = legal_a
                            action_match_found = True
                            break

                # 3. IF NOT EITHER, FIND BEST MATCH OVER THRESHOLD
                if (cross_reference is False) and (action_match_found is False):
                    # Encode the proposed action and legal actions using miniLMv6
                    if (len(action)>0) and (action.lower() not in self.null_actions):
                        action_embedding = self.encoder.encode(action)[0]
                        legal_action_embeddings = [self.encoder.encode(str(a))[0] for a in legal_actions]
                        
                        # Calculate cosine similarity between proposed action and legal actions
                        best_match_sim = 0
                        best_match_idx = 0
                        for idx, legal_emb in enumerate(legal_action_embeddings):
                            sim = self.cos(action_embedding,legal_emb).item()
                            if sim > best_match_sim:
                                best_match_idx = idx
                                best_match_sim = sim
                                            
                        if best_match_sim > 0.6:
                            # Accept the closest matching legal action
                            action = legal_actions[best_match_idx]
                            logger.info(f"Found similar legal action: {action} with similarity {best_match_sim}")
                        else:
                            print(f"Best match {legal_actions[best_match_idx]} not similar enough to action {action} with sim score {best_match_sim}.")
                            # No close match found, use random action
                            action = random.choice(legal_actions)
                            logger.warning(f"No similar legal actions found, using random choice: {action}")
                    else:
                        print(f"Null action taken: {action} - using random instead.")
                        action = random.choice(legal_actions)
            
            # Convert action back to language if action language mapping is enabled
            if self.action_language_mapping:
                action = self.action_language_mapping_decoder[action]

        return action

    # We now break agent into a policy choice, action is taken in game_env then next state is used in learning function
    def learn(self, state: str, next_state: str, r_p: float, action_code: str) -> str:
        """Given action is taken, agent learns from outcome (i.e. next state), states and actions must be text strings for LLM input."""

        # Collect outcomes in current path until a significant reward found, then use trajectory to this reward to update the knowledge
        # - Much lower costs compared to calling LLM after every step

        # Track reward history in a circular buffer
        if not hasattr(self, '_reward_array'):
            self._reward_array = np.zeros(1000)  # Pre-allocate fixed size array
            self._reward_idx = 0
        self._reward_array[self._reward_idx] = r_p
        self._reward_idx = (self._reward_idx + 1) % 1000  # Circular buffer
        self.reward_history = self._reward_array[:self._reward_idx] if self._reward_idx > 0 else self._reward_array  # View into array
    
        # Set fixed threshold for reward importance
        if abs(r_p)<=0.1:
            reward_threshold = np.max(self.reward_history)+100
        elif len(self.reward_history) > 100:
            # Calculate reward statistics over recent history
            reward_mean = np.mean(self.reward_history)
            reward_std = np.std(self.reward_history)
            
            # Define significant rewards as those outside 1 standard deviation from mean
            reward_threshold = reward_mean + reward_std
        else:
            # Dont allow it to learn until we know reward is significant
            # This will only be for the first 10 actions in the entire experiment
            reward_threshold = np.max(self.reward_history)+100

        if abs(r_p)>reward_threshold:
            print("\n SIGNIFICANT REWARD OBTAINED", r_p)
            current_diary = self._LLM_prompt_adjustment(state_action_history=self.state_action_history_current)
            self.state_action_history_current = ''
        else:
            current_outcome = 'You were positioned at' + str(state) + ', after taking action ' + str(action_code) + ' the outcome position was ' + str(next_state) + ' with reward: ' + str(r_p) +'. '
            self.state_action_history_current+=current_outcome
            # Output current knowledge from trajectory until LLM is used
            current_diary = current_outcome 
        return current_diary



import os
import json
# ------ Interaction Protocol -----------------------------------
from elsciRL.interaction_loops.standard import StandardInteractionLoop
# ------ Experiment Import --------------------------------------
from elsciRL.evaluation.standard_report import Evaluation
# ------ Gym Experiment ----------------------------------------
from elsciRL.experiments.GymExperiment import GymExperiment
# ---------------------------------------------------------------
from elsciRL.experiments.experiment_utils.agent_factory import AgentFactory
from elsciRL.experiments.experiment_utils.config_utils import ensure_dir, merge_configs
from elsciRL.experiments.experiment_utils.env_manager import EnvManager
from elsciRL.experiments.experiment_utils.result_manager import ResultManager
from elsciRL.experiments.training_procedures.default_exp_training import run_training_loop

# This is the main run functions for elsciRL to be imported
# Defines the train/test operators and imports all the required agents and experiment functions ready to be used
# The local main.py file defines the [adapters, configs, environment] to be input

# This should be where the environment is initialised and then episode_loop (or train/test) is run
# -> results then passed down to experiment to produce visual reporting (staticmethod)
# -> instruction following approach then becomes alternative form of this file to be called instead
# -> DONE: This means we have multiple main.py types (e.g. with/without convergence measure) so should create a directory and finalize naming for this

class Experiment:
    """Standard RL experiment setup, modularized."""
    def __init__(self, Config:dict, ProblemConfig:dict, Engine, Adapters:dict, save_dir:str, show_figures:str, window_size:float, 
                 training_render:bool=False, training_render_save_dir:str=None):
        # Environment setup
        if isinstance(Engine, dict):
            print("\n Multiple Engines detected, will compare results across engines...")
            self.engine_comparison = True
            self.engine_list = Engine
        else:
            self.engine_comparison = False
            self.engine_list = {'DefaultEng':Engine}
            self.engine = Engine
        self.adapters = Adapters
        self.env_manager = EnvManager(StandardInteractionLoop, Adapters)
        # ---
        # Configuration setup
        self.ExperimentConfig = Config
        self.LocalConfig = ProblemConfig
        ensure_dir(save_dir)
        self.save_dir = os.path.join(save_dir, 'Standard_Experiment')
        self.show_figures = show_figures
        if (self.show_figures.lower() != 'y') and (self.show_figures.lower() != 'yes'):
            print("Figures will not be shown and only saved.")
        self.training_render = training_render
        self.training_render_save_dir = training_render_save_dir
        try:
            self.setup_info = self.ExperimentConfig['data'] | self.LocalConfig['data'] 
        except Exception:
            self.setup_info = merge_configs(self.ExperimentConfig, self.LocalConfig)
        # Adapter/agent matching
        if 'adapter_input_dict' in self.ExperimentConfig:
            self.setup_info['adapter_input_dict'] = self.ExperimentConfig['adapter_input_dict']
        else:
            if ('adapter_select' in self.ExperimentConfig)&(len(self.ExperimentConfig['adapter_select']) > 0):
                selected_adapters = self.ExperimentConfig['adapter_select']
            else:
                selected_adapters = list(Adapters.keys())
            selected_agents = self.ExperimentConfig.get('agent_select', [])
            agent_adapter_dict = {agent_name: list(selected_adapters) for agent_name in selected_agents} if selected_agents else {}
            self.ExperimentConfig['adapter_input_dict'] = agent_adapter_dict
            self.setup_info['adapter_input_dict'] = agent_adapter_dict
        self.training_setups = {}
        self.trained_agents = {}
        self.num_training_seeds = self.setup_info.get('number_training_seeds', 1)
        self.test_agent_type = self.setup_info.get('test_agent_type', 'all')
        self.analysis = Evaluation(window_size=window_size)
        self.result_manager = ResultManager(self.analysis)
        # Gym setup
        self.is_gym_agent = {}
        for agent_type in self.setup_info.get('agent_select', []):
            if agent_type.split('_')[0] == "SB3":
                self.is_gym_agent[agent_type] = True 
                self.gym_exp = GymExperiment(Config=self.ExperimentConfig, ProblemConfig=self.LocalConfig, 
                        Engine=self.engine, Adapters=self.adapters,
                        save_dir=self.save_dir, show_figures = self.show_figures, window_size=0.1)
            else:
                self.is_gym_agent[agent_type] = False 
        self.agent_factory = AgentFactory(self.adapters, self.setup_info)

    def add_agent(self, agent_name:str, agent):
        self.agent_factory.register_agent(agent_name, agent)
        print("\n Agent added to experiment, all available agents: ", self.agent_factory.agent_types)


    def train(self):
        ensure_dir(self.save_dir)
        for engine_name, engine in self.engine_list.items():
            for agent_type in self.setup_info['agent_select']:
                is_gym_agent = self.is_gym_agent[agent_type]
                agent_adapter = (
                    agent_type + '_' + self.setup_info['adapter_input_dict'][agent_type][0]
                )
                if is_gym_agent:
                    train_setup_info = self.setup_info.copy()
                    for adapter in train_setup_info['adapter_input_dict'][agent_type]:
                        self.gym_exp.setup_info['agent_select'] = [agent_type]
                        self.training_setups[agent_adapter] = self.gym_exp.train()
                else:
                    train_setup_info = self.setup_info.copy()
                    for adapter in train_setup_info['adapter_input_dict'][agent_type]:
                        agent_parameters = train_setup_info['agent_parameters'][agent_type]
                        train_setup_info['agent_type'] = agent_type
                        train_setup_info['agent_name'] = (
                            f"{engine_name}{agent_type}_{adapter}_{agent_parameters}"
                        )
                        train_setup_info['adapter_select'] = adapter
                        train_setup_info['train'] = True
                        # Neural agent input size
                        player = self.agent_factory.create(agent_type, agent_parameters, engine, adapter)
                        train_setup_info['agent'] = player
                        train_setup_info['live_env'] = True
                        number_training_repeats = self.ExperimentConfig["number_training_repeats"]
                        
                        # Use the modular training procedure
                        self.trained_agents,_,_,_,_ = run_training_loop(
                            self.env_manager,
                            self.agent_factory,
                            self.result_manager,
                            self.training_render,
                            self.training_render_save_dir,
                            self.save_dir,
                            engine_name,
                            engine,
                            agent_type,
                            adapter,
                            self.adapters,
                            train_setup_info,
                            self.trained_agents,
                            self.num_training_seeds,
                            self.test_agent_type,
                            self.show_figures,
                            number_training_repeats
                        )
                        self.training_setups[f'Training_Setup_{engine_name}_{agent_type}_{adapter}'] = train_setup_info.copy()
        self.result_manager.training_variance_report(self.save_dir, self.show_figures)
        return self.training_setups

    # TESTING PLAY
    def test(self, training_setups:str=None):
        # Use saved training setups if not provided
        if training_setups is None:
            training_setups = self.training_setups
        else:
            training_setups = json.load(training_setups)
        for training_key in list(training_setups.keys()):
            test_setup_info = training_setups[training_key].copy()
            test_setup_info['train'] = False
            test_setup_info['training_results'] = False
            test_setup_info['observed_states'] = False
            agent_type = test_setup_info['agent_type']
            print("----------\n" + training_key)
            print("Testing results for trained agents in saved setup configuration:")
            print("TESTING SETUP INFO")
            print(test_setup_info['agent_type'])
            print(test_setup_info['adapter_select'])
            print("----------")
            agent_adapter = agent_type + "_" + test_setup_info['adapter_select']
            if self.is_gym_agent[agent_type]:
                gym_test_exp = self.training_setups[agent_adapter]
                gym_test_exp.reward_signal = None
                gym_test_exp.test()
            else:
                for engine_name, engine in self.engine_list.items():
                    for testing_repeat in range(0, test_setup_info['number_test_repeats']):
                        env = self.env_manager.create_env(engine, self.adapters, test_setup_info)
                        start_obs = env.start_obs
                        goal = str(start_obs).split(".")[0] + "---GOAL"
                        print("Flat agent Goal: ", goal)
                        agent_key = f"{engine_name}_{test_setup_info['agent_type']}_{test_setup_info['adapter_select']}"
                        if goal in self.trained_agents[agent_key]:
                            print("Trained agents available for testing.")
                            all_agents = self.trained_agents[agent_key][goal]
                            if not isinstance(all_agents, list):
                                all_agents = [all_agents]
                        else:
                            print("NO agent available for testing position.")
                            all_agents = []
                        for agent_num, agent in enumerate(all_agents):
                            env.results.reset()
                            env.start_obs = start_obs
                            env.agent = agent
                            env.agent.epsilon = 0
                            agent_adapter = test_setup_info['agent_type'] + "_" + test_setup_info['adapter_select']
                            testing_results = env.episode_loop()
                            test_save_dir = os.path.join(
                                self.save_dir,
                                f"{engine_name}_{agent_adapter}__testing_results_{str(goal).split('/')[0]}_agent{agent_num}-repeat{testing_repeat}"
                            )
                            ensure_dir(test_save_dir)
                            Return = self.result_manager.test_report(testing_results, test_save_dir, self.show_figures)
        self.result_manager.testing_variance_report(self.save_dir, self.show_figures)

        
    def render_results(self, training_setups:str=None):
        if training_setups is None:
            training_setups = self.training_setups
        else:
            training_setups = json.load(training_setups)
        for training_key in list(training_setups.keys()):
            test_setup_info = training_setups[training_key].copy()
            test_setup_info['train'] = False
            test_setup_info['training_results'] = False
            test_setup_info['observed_states'] = False
            test_setup_info['num_test_episodes'] = 1
            print("----------\nRendering trained agent's policy:")
            for engine_name, engine in self.engine_list.items():
                env = self.env_manager.create_env(engine, self.adapters, test_setup_info)
                start_obs = env.start_obs
                goal = str(start_obs).split(".")[0] + "---GOAL"
                print("Flat agent Goal: ", goal)
                agent_key = f"{engine_name}_{test_setup_info['agent_type']}_{test_setup_info['adapter_select']}"
                if goal in self.trained_agents[agent_key]:
                    print("Trained agents available for testing.")
                    all_agents = self.trained_agents[agent_key][goal]
                    if not isinstance(all_agents, list):
                        all_agents = [all_agents]
                else:
                    print("NO agent available for testing position.")
                    all_agents = []
                for agent_num, agent in enumerate(all_agents):
                    env.results.reset()
                    env.start_obs = start_obs
                    env.agent = agent
                    env.num_train_repeat = 1
                    env.num_test_repeat = 1
                    env.number_episodes = 1
                    env.agent.epsilon = 0
                    render_save_dir = os.path.join(self.save_dir, 'render_results')
                    ensure_dir(render_save_dir)
                    render_results = env.episode_loop(render=True, render_save_dir=render_save_dir)

        return render_results


import torch
from typing import List, Any
from torch import Tensor
from tqdm import tqdm
from elsciRL.encoders.encoder_abstract import StateEncoder

class StateEncoder(StateEncoder):
    def __init__(self, num_states):
        """Encoder for default state representation produced by the environment/engine."""
        # Create dict lookup
        # - get binary list that indexes the state e.g. 0_0 -> [1,0,0,0] or 0_3 -> [0,0,0,1]
        # UPDATED - Now uses torch.nn.functional.one_hot for one-hot encoding
        # Using one-hot encoder is incredibly inefficient for large state spaces
        # Instead, we consider using an index-based encoding where each unique state is assigned a unique index.
        self.device = "cuda" if torch.cuda.is_available() else "cpu" # Make this optional choice with parameter
        self.vectors: Tensor = torch.cat([torch.eye(num_states), torch.zeros(1,num_states)]).to(self.device)         # tensor needs to be defined to len(local_object)
        self.name = "StateEncoder"
        self.input_type = "list"
        self.output_type = "tensor"
        self.output_dim = num_states

        self.encoder = {}
        self.encoder_idx = 0
        self.num_states = num_states

    def encode(self, state:Any = None, legal_actions:list = None, episode_action_history:list = None,
               indexed: bool = False) -> Tensor:
        """ Set of all possible states are simply converted to a vector"""
        # One hot encode the state if it is not already indexed
        if state not in self.encoder:
            state_encoded = self.encoder_idx  # Use the index as the state encoded value
            # Store the encoded state in the encoder dictionary
            self.encoder[state] = state_encoded
            # Increment the encoder index for the next unique state
            self.encoder_idx += 1
        else:
            state_encoded = self.encoder[state]

        # If indexed, use one-hot encoding
        # If not indexed, use the unique index to retrieve the vector
        if indexed:
            state_encoded = torch.nn.functional.one_hot(torch.tensor(state_encoded), num_classes=self.num_states).float().to(self.device)
        else:
            state_encoded = self.vectors[int(state_encoded)].flatten()
        
        return state_encoded    